/*
!C******************************************************************

!Description: 
	Functions for writing information to be used in generating
	a MODIS-like VIIRS fire Climate Modeling Grid (CMG) product.

!Revision History:
	Original MODIS version, April 2002, L. Giglio.
	Revised June 2006 to separate day and night observations.
	Revised November 2010 to handle processing of water pixels in
	Collection 6 detection algorithm.
	Adapted to VIIRS, April 2012, L. Giglio.
	Revised September 2013 to match version 6.2.1 of MODIS fire code.

!Design Notes

!END*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "mfhdf.h"
#include "const.h"
#include "types.h"
#include "cmg.h"
#include "error.h"
#include "fire.h"



/* 0.25 degree bins covering the entire globe */
#define CMG_BIN_SIZE 0.25F
#define CMG_ROWS 720
#define CMG_COLS 1440
#define CMG_ELEMS (CMG_ROWS * CMG_COLS)



typedef struct {
	USHORT total, water;
	USHORT land_missing, land_cloud, land_unknown, land_fire;
	} GridCell;



/*
CMG fields that will be written to an SDS in the HDF output file.

CMG_VALUES must always be last enumeration so the code can use it
to determine the total number of fields to be written.
*/
enum {CMG_X, CMG_Y, CMG_TOTAL, CMG_WATER,
	CMG_LAND_MISSING, CMG_LAND_CLOUD, CMG_LAND_UNKNOWN,
	CMG_LAND_FIRE, CMG_VALUES};


static GridCell *cmg_day = (GridCell *) NULL;
static GridCell *cmg_night = (GridCell *) NULL;


extern int verbose;


int cmg_init(int flag)
{
	if (!flag) return 0;

	if ( !(cmg_day = calloc(CMG_ELEMS, sizeof(GridCell))) )
		return error(ERR_F_MEM);
	if ( !(cmg_night = calloc(CMG_ELEMS, sizeof(GridCell))) )
		return error(ERR_F_MEM);

	return 0;
}


/* check for bad geolocation before calling this function */
void cmg_row_col(float lat, float lon, int *row, int *col)
{
	*row = ((90.0F - lat) / CMG_BIN_SIZE);
	*col = ((lon + 180.0F) / CMG_BIN_SIZE);

	/* handles legitimate case when longitude = 180E, but also tacitly
	clamps bogus longitudes above 180E to right column (a side effect) */
	if (*row >= CMG_ROWS) *row = CMG_ROWS - 1;
	if (*col >= CMG_COLS) *col = CMG_COLS - 1;
}


void cmg_append(float lat, float lon, uint8 emaskval, PixInfo pixflags)
{
	int row, col;
	long i;
	GridCell *cmg;


	if (pixflags.day) cmg = cmg_day; else cmg = cmg_night;

	/* exclude missing or bad geolocation */
	if (pixflags.missing_geo) return;

	cmg_row_col(lat, lon, &row, &col);

	i = row * CMG_COLS + col;

	/* will never happen, but just in case we'll tally bad points in
	the otherwise-unused upper left grid cell */
	if (i < 0L) i = 0L;

	cmg[i].total++;

	switch (emaskval) {
		case ECLASS_CLEAR_WATER:
		case ECLASS_CLOUDY_WATER:
		case ECLASS_UNKNOWN_WATER:
		case ECLASS_FIRE_WATER:
			cmg[i].water++;
			break;

		case ECLASS_CLOUDY_LAND:
			cmg[i].land_cloud++;
			break;

		case ECLASS_UNKNOWN_LAND:
			cmg[i].land_unknown++;
			break;

		/* NB. missing geolocation handled above */
		case ECLASS_MISS:
			if (pixflags.surf == LAND)
				cmg[i].land_missing++;
			else
				cmg[i].water++;
			break;

		case ECLASS_FIRE_LAND:
			cmg[i].land_fire++;
			break;
		}
}





void cmg_free(int flag)
{
	if (flag) {
		if (cmg_day) free(cmg_day);
		if (cmg_night) free(cmg_night);

		/* ensure these are null for debugging and to prevent problems
		if cmg_free() is mistakenly called more than once */
		cmg_day = cmg_night = (GridCell *) NULL;
		}
}



/* keep this at or above ~1000 otherwise HDF internal compression will
become inefficient */
#define BUFF_ROWS 2000



int cmg_write(int32 sd_id, int day)
/*
!C*****************************************************************************

!Description:
	Create and writes the SDS's comprising data layers that will
	be used to create the fire CMG product from the L2 fire product.

!Input Parameters:
	sd_id	HDF SD ID.
	day	flag: non-zero = write daytime SDS, otherwise write nighttime

!Output Parameters:
	None.

!Input/Ouput Parameters:
	None.

!Returns:
	Returns 0 if successful, non-zero otherwise.

!Revision History:
	Original version, L. Giglio (April 2002)
	Revised June 2006 to handle separate day and night observations and
	enable internal HDF compression.

!Design Notes

!END
****************************************************************************/
{
	int i, k;
	int32 sds_id;
	int32 dim_sizes[2];
	int32 start[2], edges[2];
	int32 dim_id;

	int ncells;
	GridCell *cmg;
	char *dimname, *sdsname;

	HDF_CHUNK_DEF cdef;
	uint16 buff[BUFF_ROWS][CMG_VALUES];



	if (day) {
		cmg = cmg_day;
		dimname = "cmg_cells_day";
		sdsname = "CMG_day";
		}
	else {
		cmg = cmg_night;
		dimname = "cmg_cells_night";
		sdsname = "CMG_night";
		}


	/* count number of nonzero cells */
	ncells = 0;
	for (i = 0; i < CMG_ELEMS; i++)
		if (cmg[i].total) ncells++;

if (verbose) printf("%d %s cells\n", ncells, sdsname);


	/* leave if there are no cells to write */
	if (!ncells) return 0;


	dim_sizes[0] = ncells;
	dim_sizes[1] = CMG_VALUES;

	if ((sds_id = SDcreate(sd_id, sdsname, DFNT_UINT16, 2, dim_sizes))
		== FAIL) return error(ERR_HDF);

	/* chunk and compress CMG data only if a substantial number of
	values are to be written */
	if (ncells >= 200) {
		cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = BUFF_ROWS;
		cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = CMG_VALUES;
		cdef.comp.comp_type = COMP_CODE_DEFLATE;
		cdef.comp.cinfo.deflate.level = 8;

		if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
			return error(ERR_HDF);
		}

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
		return error(ERR_HDF);
	if (SDsetdimname(dim_id, dimname) == FAIL)
		return error(ERR_HDF);

	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL)
		return error(ERR_HDF);
	if (SDsetdimname(dim_id, "cmg_values") == FAIL)
		return error(ERR_HDF);


	start[1] = 0;
	edges[1] = CMG_VALUES;

	start[0] = 0;
	edges[0] = BUFF_ROWS;
	k = 0;
	for (i = 0; i < CMG_ELEMS; i++) {
		if (!cmg[i].total) continue;

		if (k >= BUFF_ROWS) {
			if (SDwritedata(sds_id, start, NULL, edges, buff) == FAIL)
				return error(ERR_HDF);
			k = 0;
			start[0] += BUFF_ROWS;
			}

		buff[k][CMG_X] = i % CMG_COLS;
		buff[k][CMG_Y] = i / CMG_COLS;

		buff[k][CMG_TOTAL] = cmg[i].total;
		buff[k][CMG_WATER] = cmg[i].water;

		buff[k][CMG_LAND_MISSING] = cmg[i].land_missing;
		buff[k][CMG_LAND_CLOUD] = cmg[i].land_cloud;
		buff[k][CMG_LAND_UNKNOWN] = cmg[i].land_unknown;
		buff[k][CMG_LAND_FIRE] = cmg[i].land_fire;

		k++;
		}

	/* write partial buffer */
	if (k) {
		edges[0] = k;
		if (SDwritedata(sds_id, start, NULL, edges, buff) == FAIL)
			return error(ERR_HDF);
		}

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	return 0;
}


