#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "confidence.h"


static float S(float x, float a, float b);



int confidence(float T13, float z13, float DT, float MeanDT, float MadDT,
	int AdjCloud, int AdjWater, int day, float thresh_PF_T13)
/*
!C*****************************************************************************
!Description:
	Experimental function to compute confidence of detected
	fire pixels.

!Input Parameters:
	T13
	z13
	DT
	MeanDT
	MadDT
	AdjCloud
	AdjWater	number of water pixels adjacent to fire pixel
	day		day/night flag (non-zero --> day)
	thresh_PF_T21

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Returns detection confidence in percent.

!Revision History
	Original version April 2001, L. Giglio.

	L. Giglio, Jul. 2004
	Properly handle case of failed background characterization.

	L. Giglio, April 2006
	Slightly altered daytime thresholds for C1, C4, and C5.

	L. Giglio, August 2011
	Dropped adjacent cloud criterion and altered thresholds for
	Collection 6 algorithm.

        L. Giglio, August 2011 and July 2013
        Altered tests based on MODIS Collection 6 algorithm.

!END***************************************************************************
*/
{
	float C1, C2, C3, C4, C5;	/* individual confidence tests */
	float C;
	float zDT;


	/* tiny value used to prevent division by zero */
	const float eps = 1.0e-10F;


	/* special case: no background statistics available */
	if (z13 < eps) {
		/*
		No background statistics are available.  However, we know that
		the conservative absolute threshold test was passed, so T13
		will be at least 360 K (= DayParams.thresh_T13) during the day
		and 320 K (= NightParams.thresh_T13) at night.  Because these
		thresholds are so conservative, we choose the ramp function
		parameters so that pixels meeting even these minimum thresholds
		will have a confidence of 90%, and rapidly increase to 100% for
		brightness temperatures above these thresholds.
		*/
		if (day)
			C = S(T13, 270.0F, 370.0F);
		else
			C = S(T13, 230.0F, 330.0F);

		/* convert to percent */
		return (int) (100.0F * C);
		}


	zDT = (DT - MeanDT) / (MadDT + eps);

	/* keep these consistent with algorithm parameters */
	C2 = S(z13, 3.0F, 6.0F);
	C3 = S(zDT, 3.5F, 6.0F);


	if (day) {
		C1 = S(T13, thresh_PF_T13, 360.0F);
		C4 = 1.0F - S((float) AdjCloud, 0.0F, 4.0F);
		C5 = 1.0F - S((float) AdjWater, 0.0F, 4.0F);
		C = powf(C1 * C2 * C3 * C4 * C5, 0.2F);
		}
	else {
		/* night */
		C1 = S(T13, thresh_PF_T13, 320.0F);
		C = powf(C1 * C2 * C3, 0.333333F);
		}

	/* convert to percent */
	return (int) (100.0F * C);
}





/*
Evaluate unit ramp function having specified thresholds.
Returns:
	0		(x < a)
	(x-a)/(b-a)	(a <= x <= b)
	1		(x > b)
*/
static float S(float x, float a, float b)
{
	if (x <= a) return 0.0F;
	if (x >= b) return 1.0F;

	return (x - a) / (b - a);
}


