#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
#include "fire.h"
#include "land_cover.h"
#include "detfire.h"
#include "firetab.h"
#include "cloud.h"
#include "confidence.h"
#include "frp.h"
#include "zstats.h"
#include "cmg.h"
#include "qf.h"


#undef max
#define max(a,b) (a < b ? b : a)


#define MAX_WIN_SIZE 21
#define MAX_WIN_PIX (MAX_WIN_SIZE * MAX_WIN_SIZE)


typedef struct {
	/* thresholds used to identify potential fire pixels */
	float	thresh_PF_T13;	/* K */
	float	thresh_PF_DT;	/* K */
	float	thresh_PF_R7;	/* near-IR reflectance threshold, daytime algorithm only */

	/* thresholds used to identify valid bkg. pixels */
	float	thresh_bkg_T13;	/* K */
	float	thresh_bkg_DT;	/* K */

	float	thresh_T13;	/* K */

	float	minbkg_DT;	/* K; minimum difference above bkg. DT */

	float	devrp_T15;	/* K; T15 absolute deviation reduction parameter (daytime only) */
	} AlgParams;


static int test_potential_fire_pixel(AlgParams *ap, int i, int j,
	int first_line, Scan *PrevScan, Scan *scan, Scan *NextScan,
	ScanThreshPF *pfthresh, uint8 *FireMask, uint8 *ExtFireMask,
	uint32 *QA, FireCounts *counts);

static int BkgPixels(int ic, int jc, int surf,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT13, float *BuffT15, float *BuffDT, float *BuffL21,
	float *BuffR5, float *BuffR7, float *BuffR11, int *nref,
	float *BkgFireBuffT13, float *BkgFireBuffT15,
	float *BkgFireBuffDT, int *nbfire, int *winparam,
	int *nwater, int *nland, int *ncoast);

static void GetValidBkgPixels(int ic, int jc, int r, int surf_state,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT13, float *BuffT15, float *BuffDT, float *BuffL21,
	int *nvalid,
	float *BuffR5, float *BuffR7, float *BuffR11, int *nref,
	float *BkgFireBuffT13, float *BkgFireBuffT15, float *BkgFireBuffDT,
	int *nbfire, int *nwater, int *nland, int *ncoast);

static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater);

static void compute_stats(float *x, int n, float *mean, float *mad);

static int SunGlint(float, float, float, float, float, float, float *);
static float fixang(float);

static void compute_ndvi(float *R5, float *R7, int n, float *NDVI);
static int bkg_water_pix(float *R7, float *R11, float *NDVI, int n);

static float calc_thresh_PF_T13(float zone_mean_T13);
static float calc_thresh_PF_DT(float zone_mean_DT);


static AlgParams DayParams = {
	310.0F,	10.0F,	/* no fire */
	0.35F,		/* max. band 7 reflectance for fire pixels */

	325.0F,	20.0F,	/* valid bkg. */

	360.0F,		/* absolute threshold test */

	6.0F,

	4.0F
	};


static AlgParams NightParams = {
	305.0F,	10.0F,	/* no fire */
	0.0F,		/* dummy band 7 "nighttime reflectance" */

	310.0F,	10.0F,	/* valid bkg. */

	320.0F,		/* absolute threshold test */

	6.0F,

	1000.0F		/* dummy nighttime value that will cause test to always be passed */
	};




void prescreen(Scan *scan, int skip_water)
/*
!C*****************************************************************************
!Description:
	Pre-screen current buffer prior to performing detection algorithm.
	Sets individual flags within the "pixinfo" field of the Scan
	structure pointed to by "scan".
	This function will also swap contents of individual elements of
	scan->T13 and scan->T22 arrays as necessary to obtain valid,
	unsaturated mid-infrared brightness temperatures to process.

!Input Parameters:
	skip_water	flag: process water pixels
			(For the C6 algorithm this flag is normally set.
			DB ground stations may wish to skip water pixels
			when processing, however, so this flag is
			included in the code.)
!Output Parameters:

!Input/Output Parameters:
	scan		pointer to current scan buffer

!Returns:
	Nothing.

!Revision History

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;
	int surf;
	int processed_water;	/* flag: water pixel that will be processed */
	int process;		/* flag: land or processed-water pixel */
	int day;		/* flag: daytime pixel */
	int bad_geo;		/* flag: invalid geolocation for pixel */
	int missing_therm;	/* flag: one or more missing thermal bands */
	int missing_ref;	/* flag: one or more missing reflective bands */

	int glint;		/* flag: glint = non-zero */
	float GlintAng;		/* degrees */

	int zflag;		/* flag: include pixel when computing
				zone statistics */

	float T13, T15, T16;
	float R5, R7, R11;
	int bad13, bad15, bad16;
	float RelAzAng;
	AlgParams *ap;


	PixInfo NullPixInfo;


	(void) memset(&NullPixInfo, 0, sizeof(PixInfo));

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			/* clear internal flags */
			scan->pixflags[i][j] = NullPixInfo;

			/*
 			Use QST LW mask to identify trimmed EDR pixels.
			EDRs have additional trimming beyond the on-board
			SDR trimming.  No additional QA bit fields will be set
			for trimmed pixels.
			*/
			if (lw_trim(scan->landsea[i][j])) {
				scan->pixflags[i][j].trim = 1;
				continue;
				}

			T13 = scan->T13[i][j];
			T15 = scan->T15[i][j];
			T16 = scan->T16[i][j];

			R5 = scan->R5[i][j];
			R7 = scan->R7[i][j];
			R11 = scan->R11[i][j];

			bad13 = test_therm_QF(13, scan->QF13[i][j], T13);
			bad15 = test_therm_QF(15, scan->QF15[i][j], T15);
			bad16 = test_therm_QF(16, scan->QF16[i][j], T16);


			/* check that we have valid geolocation */
			bad_geo = ((fabsf(scan->lat[i][j]) > 90.0) ||
				(fabsf(scan->lon[i][j]) > 180.0));
			if (bad_geo) {
				scan->pixflags[i][j].missing_geo = 1;
				continue;
				}

			/* set day/night flag */
			day = scan->pixflags[i][j].day =
				scan->SolZenAng[i][j] < 85.0F;


			if (lw_missing(scan->landsea[i][j])) {
				scan->pixflags[i][j].missing_lc = 1;
				continue;
				}

			if (lw_land(scan->landsea[i][j])) {
				surf = scan->pixflags[i][j].surf = LAND;
				processed_water = 0;
				}
			else {
				/* water or coast */
				if (lw_coast(scan->landsea[i][j]))
					surf = COAST;
				else
					surf = WATER;

				scan->pixflags[i][j].surf = surf;

				if (skip_water)
					processed_water = 0;
				else {
					/*
					Omit coast and inland water from the
					fire detection algorithm.
					*/
					processed_water =
						lw_ocean(scan->landsea[i][j]);
					}
				}


			/*
			Check availability of thermal bands.  We do this
			AFTER setting the day/night and land/water flags so
			that the CMG module can properly tally missing pixels.
			*/
			missing_therm = bad13 || bad15 || bad16;

			if (missing_therm) {
				scan->pixflags[i][j].missing_rad = 1;
				continue;
				}


			/*
 			This test may ultimately need to be tweaked to
			consider the reflective-band QF values.
 			*/
			missing_ref = (R5 <= 0.0F) || (R7 <= 0.0F) ||
				(R11 <= 0.0F);

			if (day && missing_ref) {
				/* at least for now flag missing reflectance
 				using the missing radiance flag since the
				algorithm needs both to proceed */
				scan->pixflags[i][j].missing_rad = 1;
				continue;
				}



			if (day) {
				/* sun-glint test */
				RelAzAng = fixang(scan->SolAzAng[i][j] -
					scan->ViewAzAng[i][j]);

				glint = SunGlint(R5, R7, R11,
					scan->SolZenAng[i][j],
					scan->ViewZenAng[i][j],
					RelAzAng, &GlintAng);

				scan->pixflags[i][j].glint_level = glint;
				}
			else
				glint = 0;


			/* run cloud mask on all pixels */
			if (T16 > 0.0F) {
				if (iscloud(R5, R7, T16, surf == WATER)) {
					scan->pixflags[i][j].cloud = 1;
					continue;
					}
				}
/* consider setting missing cloud mask flag */

			/*
			At this point we have a cloud-free pixel.
			Detection algorithm will be applied to
			all land pixels, no coastal pixels, and only a subset
			of water pixels.
			*/

			scan->pixflags[i][j].process = process =
				(surf == LAND) || processed_water;


			/* stop processing current pixel if we're not going
			to apply the detection algorithm to it */
			if (!process) continue;

			/* at this point we have either a land pixel or
			a processed-water pixel */

			/* identify the subset of cloud-free pixels that are
			"safe" to use when computing zone statistics */
			if (day)
				ap = &DayParams;
			else
				ap = &NightParams;

/* FIX: may want to use more stringent glint test over water */
			zflag = (T13 <= ap->thresh_T13) && (glint <= 1);

			scan->pixflags[i][j].zsafe = zflag;
			}
		}
}




/*
Calculate VIIRS band 13 potential fire threshold given the mean T13 value
for a zone.
*/
static float calc_thresh_PF_T13(float zone_mean_T13)
{
	float thresh_T13;
	const float min_thresh_T13 = 300.0F;
	const float max_thresh_T13 = 330.0F;

	thresh_T13 = zone_mean_T13 + 5.0F;

	if (thresh_T13 < min_thresh_T13) return min_thresh_T13;
	if (thresh_T13 > max_thresh_T13) return max_thresh_T13;

	return thresh_T13;
}



static float calc_thresh_PF_DT(float zone_mean_DT)
{
	float thresh_DT;
	const float min_thresh_DT = 10.0F;
	const float max_thresh_DT = 35.0F;

	thresh_DT = zone_mean_DT + 5.0F;

	if (thresh_DT < min_thresh_DT) return min_thresh_DT;
	if (thresh_DT > max_thresh_DT) return max_thresh_DT;

	return thresh_DT;
}




/* compute potential fire thresholds for each zone */
void compute_scan_PF_thresholds(Scan *scan, ScanThreshPF *pfthresh)
{
	int i, j;
	AlgParams *ap;

	float thresh_PF_T13, thresh_PF_DT;	/* kelvins */
	const float missval = -1.0F;	/* use negative value for missing */

	static float land_thresh_PF_T13[SAMPLES_750M];
	static float land_thresh_PF_DT[SAMPLES_750M];


	for (j = 0; j < SAMPLES_750M; j++) {
		if (scan->LandZonePix[j] >= MIN_VALID_ZONE_PIX) {
			land_thresh_PF_T13[j] =
				calc_thresh_PF_T13(scan->LandMeanT13[j]);
			land_thresh_PF_DT[j] =
				calc_thresh_PF_DT(scan->LandMeanDT[j]);
			}
		else {
			land_thresh_PF_T13[j] = missval;
			land_thresh_PF_DT[j] = missval;
			}
		}


	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (!scan->pixflags[i][j].process) {
				pfthresh->thresh_PF_T13[i][j] = missval;
				pfthresh->thresh_PF_DT[i][j] = missval;
				continue;
				}

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;

			if (scan->pixflags[i][j].surf == LAND) {
				thresh_PF_T13 = land_thresh_PF_T13[j];
				thresh_PF_DT = land_thresh_PF_DT[j];
				}
			else {
				/* force use of default thresholds over water */
				thresh_PF_T13 = thresh_PF_DT = missval;
				}

			/*
			Use conservative fixed potential-fire thresholds if
			sample statistics are unavailable, or if we are
			processing a water pixel.
			*/
			if (thresh_PF_T13 < 0.0F) {
				thresh_PF_T13 = ap->thresh_PF_T13;
				thresh_PF_DT = ap->thresh_PF_DT;
				}

			pfthresh->thresh_PF_T13[i][j] = thresh_PF_T13;
			pfthresh->thresh_PF_DT[i][j] = thresh_PF_DT;
			}
		}
}





/*
Flag potential fire pixels in specified scan.

Functions prescreen(), compute_zone_stats(), and compute_zone_PF_thresholds()
must be called prior to calling this function.
*/
void flag_potential_fires(Scan *scan, ScanThreshPF *pfthresh)
{
	int i, j;
	float T13, T15, DT;			/* kelvins */
	AlgParams *ap;
	int pftest1, pftest2, pftest3;

	float thresh_PF_T13, thresh_PF_DT;	/* kelvins */


	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (!scan->pixflags[i][j].process) continue;

			T13 = scan->T13[i][j];
			T15 = scan->T15[i][j];
			DT = T13 - T15;

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;


			thresh_PF_T13 = pfthresh->thresh_PF_T13[i][j];
			thresh_PF_DT = pfthresh->thresh_PF_DT[i][j];

			pftest1 = T13 > thresh_PF_T13;
			pftest2 = DT > thresh_PF_DT;
			pftest3 = scan->R7[i][j] < ap->thresh_PF_R7;

			/* potential fire test */
			if (pftest1 && pftest2 && pftest3)
				scan->pixflags[i][j].potential_fire = 1;
			}
		}
}




/*
Flag background fire pixels in specified scan.
See 2003 RSE paper for definitions and details.

Function prescreen() must be called prior to calling this function.
*/
void flag_background_fires(Scan *scan)
{
	int i, j;
	float T13, T15, DT;			/* kelvins */
	AlgParams *ap;

	int bfire;	/* flag: background fire pixel */


	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (!scan->pixflags[i][j].process) continue;

			T13 = scan->T13[i][j];
			T15 = scan->T15[i][j];
			DT = T13 - T15;

			if (scan->pixflags[i][j].day)
				ap = &DayParams;
			else
				ap = &NightParams;

			/* valid background pixel test */
			bfire = (T13 > ap->thresh_bkg_T13) &&
				(DT > ap->thresh_bkg_DT);

			scan->pixflags[i][j].bkg_fire = bfire;
			scan->pixflags[i][j].valid_bkg_pix = !bfire;
			}
		}
}




int detect(Scan *PrevScan, Scan *scan, Scan *NextScan, int first_line,
	int skip_water, ScanThreshPF *pfthresh, OutScan *outscan,
	FireCounts *counts)
/*
!C*****************************************************************************
!Description:
	Apply detection algorithm to current scan.

!Input Parameters:
	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

	first_line
	skip_water
	pfthresh

!Output Parameters:
	outscan
	counts

!Input/Output Parameters:
	None.

!Returns:
	0 if no errors, non-zero if fire pixel table functions encountered
	error allocating memory.

!Revision History
	Original version April 2001, L. Giglio.


!END***************************************************************************
*/
{
	int i, j;
	int day;		/* flag: day = non-zero, night = 0 */
	int land;
	int coast;
	int missing;
	AlgParams *ap;

	uint8 fmask, efmask;
	uint32 QA;

	int st;



	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			/* clear all QA bits */
			outscan->FireQA[i][j] = 0;
outscan->ExtFireMask[i][j] = ECLASS_BOGUS;

			if (scan->pixflags[i][j].trim) {
				outscan->FireMask[i][j] = CLASS_NOPROC_TRIM;
outscan->ExtFireMask[i][j] = ECLASS_TRIM;
				counts->NumTrimPix++;
				continue;
				}

			day = scan->pixflags[i][j].day;
			land = scan->pixflags[i][j].surf == LAND;
			coast = scan->pixflags[i][j].surf == COAST;

			/* copy land/water state into QA bits 0 and 1 */
			outscan->FireQA[i][j] |= scan->pixflags[i][j].surf;

			if (day) {
				ap = &DayParams;

				/* set day/night algorithm flag in product */
				outscan->FireQA[i][j] |= 0x10;

				counts->NumDayPix++;
				}
			else {
				ap = &NightParams;

				counts->NumNightPix++;
				}

			/* missing data? */
			missing = scan->pixflags[i][j].missing_rad ||
				scan->pixflags[i][j].missing_geo ||
				scan->pixflags[i][j].missing_lc;
			if (missing) {
				outscan->FireMask[i][j] = CLASS_NOPROC_MISS;
outscan->ExtFireMask[i][j] = ECLASS_MISS;
				counts->NumMissingPix++;
				counts->NumMissingRad +=
					scan->pixflags[i][j].missing_rad;
				counts->NumMissingGeo +=
					scan->pixflags[i][j].missing_geo;
				continue;
				}


			/* update counts of glint pixels and set glint bit */
			if (scan->pixflags[i][j].glint_level) {
				counts->NumGlintPix++;
				outscan->FireQA[i][j] |=
					(scan->pixflags[i][j].glint_level << 22);
				}


			if (land)
				counts->NumLandPix++;
			else
				counts->NumWaterPix++;

/* FIX: count coast pixels separately */


			/* mimic C5 handling of water pixels if requested */
			if (skip_water && !land) {
				outscan->FireMask[i][j] = CLASS_WATER;
				outscan->ExtFireMask[i][j] =
					ECLASS_UNKNOWN_WATER;
/* FIX: set QA? */
				continue;
				}


			/* cloud? */
			if (scan->pixflags[i][j].cloud) {
				if (land) {
					/* cloud over land */
					outscan->FireMask[i][j] = CLASS_CLOUD;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLOUDY_LAND;
					counts->NumLandCloudPix++;
					}
				else {
					/* cloud over water or coast */
					outscan->FireMask[i][j] = CLASS_CLOUD;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLOUDY_WATER;
					counts->NumWaterCloudPix++;
					}

				continue;
				}


			/* handle unprocessed clear-water pixels */
			if (!scan->pixflags[i][j].process) {
				/* coast or unprocessed water? */
				if (!land) {
					outscan->FireMask[i][j] = CLASS_WATER;
					outscan->ExtFireMask[i][j] =
						ECLASS_UNKNOWN_WATER;
					counts->NumUnknownWaterPix++;
					}
#ifdef DEBUG
				else {
					/* unprocessed land pixels should
					never have made it to this point */
puts("UNEXPECTED UNPROCESSED LAND PIXEL");
					}
#endif

				continue;
				}


			/* obvious non-fire? (== not a potential fire) */
			if ( !scan->pixflags[i][j].potential_fire ) {
				if (land) {
					outscan->FireMask[i][j] = CLASS_NO_FIRE;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLEAR_LAND;
					}
				else {
					outscan->FireMask[i][j] = CLASS_WATER;
					outscan->ExtFireMask[i][j] =
						ECLASS_CLEAR_WATER;
					}

/* FIX: this used to be land pixels only */
				counts->NumNoFirePix++;
				continue;
				}


			/* At this point we have a potential fire pixel,
			either over water or land. */

			/* set algorithm QA: potential fire pixel */
			outscan->FireQA[i][j] |= 0x20;

			st = test_potential_fire_pixel(ap, i, j, first_line,
				PrevScan, scan, NextScan, pfthresh,
				&fmask, &efmask, &QA, counts);
			if (st) return st;

			outscan->FireMask[i][j] = fmask;
			outscan->ExtFireMask[i][j] = efmask;
			outscan->FireQA[i][j] |= QA;
			}
		}

	return 0;
}






/*
Apply fire detection tests to a potential fire pixel.


Input Variables
	ap
	i
	j
	PrevScan
	scan
	NextScan

Input/Output Variables
	outscan		pointer to output scan buffer; element [i,j] of
			FireMask, ExtFireMask, and FireQA fields will be
			set by this function

	counts		pointer to FireCounts structure; fields will be
			updated as necessary

Output Variables
	FireMask
	ExtFireMask
	QA


!Returns:
	0 if no errors, non-zero if fire pixel table functions encountered
	error allocating memory.

*/
static int test_potential_fire_pixel(AlgParams *ap, int i, int j,
	int first_line, Scan *PrevScan, Scan *scan, Scan *NextScan,
	ScanThreshPF *pfthresh,
	uint8 *FireMask, uint8 *ExtFireMask, uint32 *QA, FireCounts *counts)
{
	float T13, T15;
	float DT;
	int day;		/* flag: day = non-zero, night = 0 */
	int land;

	int test1, test2, test3, test4, test5, test6;
	int fire;		/* flag: fire pixel */

	int winparam;
	int L;			/* background window size (L x L) */

	float MeanT13, MeanDT, MeanT15;
	float MadT13, MadDT, MadT15;
	float z13, z15;
	const float eps = 1.0e-10F;
	float MeanR7, MadR7;

	int AdjCloud, AdjWater;
	int C;			/* detection confidence */

	float MeanL13;
	float power;		/* FRP [MW] */
	float fdummy;

	float RelAzAng;		/* degrees */

	int st;

	int nvalid;
	float BuffT13[MAX_WIN_PIX], BuffT15[MAX_WIN_PIX], BuffDT[MAX_WIN_PIX];
	float BuffL13[MAX_WIN_PIX];
	int nref;
	float BuffR5[MAX_WIN_PIX], BuffR7[MAX_WIN_PIX], BuffR11[MAX_WIN_PIX];
	float BuffNDVI[MAX_WIN_PIX];

	/* pixels that appear to be unmasked water pixels contaminating
	the background sample over land */
	int nwater_cont;

	float BkgFireMeanT13, BkgFireMeanDT, BkgFireMeanT15;
	float BkgFireMadT13, BkgFireMadDT, BkgFireMadT15;

	int nbfire;
	float fbfire;	/* fraction of pixels w/in window that are background
			fire pixels */

	int nwater, nland, ncoast;

	int nwc; /* nwc = nwater + ncoast; equivalent to nwater in C5 code */


	float BkgFireBuffT13[MAX_WIN_PIX], BkgFireBuffT15[MAX_WIN_PIX],
		BkgFireBuffDT[MAX_WIN_PIX];

	int cmg_row, cmg_col;


	*QA = 0;


	T13 = scan->T13[i][j];
	T15 = scan->T15[i][j];
	DT = T13 - T15;
	day = scan->pixflags[i][j].day;
	land = scan->pixflags[i][j].surf == LAND;


	/* compute background statistics */
	nvalid = BkgPixels(i, j, scan->pixflags[i][j].surf,
		PrevScan, scan, NextScan,
		BuffT13, BuffT15, BuffDT, BuffL13,
		BuffR5, BuffR7, BuffR11, &nref,
		BkgFireBuffT13, BkgFireBuffT15, BkgFireBuffDT,
		&nbfire, &winparam, &nwater, &nland, &ncoast);


	if (nbfire > 0) {
		compute_stats(BkgFireBuffT13, nbfire,
			&BkgFireMeanT13, &BkgFireMadT13);
		compute_stats(BkgFireBuffDT, nbfire,
			&BkgFireMeanDT, &BkgFireMadDT);
		compute_stats(BkgFireBuffT15, nbfire,
			&BkgFireMeanT15, &BkgFireMadT15);
		}
	else {
		BkgFireMeanT13 = BkgFireMeanT15 = BkgFireMeanDT = 0.0F;
		/* this will force test 6 to fail */
		BkgFireMadT13 = BkgFireMadT15 = BkgFireMadDT = 0.0F;
		}


	if (nvalid > 0) {
		compute_stats(BuffT13, nvalid, &MeanT13, &MadT13);
		compute_stats(BuffDT, nvalid, &MeanDT, &MadDT);
		compute_stats(BuffT15, nvalid, &MeanT15, &MadT15);

		fbfire = (float) nbfire / (float) nvalid;

		z13 = (T13 - MeanT13) / (MadT13 + eps);
		z15 = (T15 - MeanT15) / (MadT15 + eps);

		test2 = DT > (MeanDT + 3.5F * MadDT);
		test3 = DT > (MeanDT + ap->minbkg_DT);
		test4 = T13 > (MeanT13 + 3.0F * MadT13);
		test5 = T15 > (MeanT15 + MadT15 - ap->devrp_T15);
		test6 = BkgFireMadT13 > 5.0F;
		}
	else {
		MeanT13 = MeanT15 = MeanDT = 0.0F;
		MadT13 = MadT15 = MadDT = 0.0F;
		z13 = z15 = 0.0F;
		test2 = test3 = test4 = test5 = test6 = 0;
		fbfire = 0.0F;
		}

	test1 = T13 > ap->thresh_T13;

	/* set background characterization field */
	*QA |= (winparam << 7);

	/* save individual test results */
	if (test1) *QA |= 0x800;	/* bit 11 */
	if (test2) *QA |= 0x1000;	/* bit 12 */
	if (test3) *QA |= 0x2000;	/* bit 13 */
	if (test4) *QA |= 0x4000;	/* bit 14 */
	if (test5) *QA |= 0x8000;	/* bit 15 */
	if (test6) *QA |= 0x10000;

	/*
	There are so many routes below by which a non-fire
	class can be assigned that we will assume this result
	by default and later modify if necessary.
	*/
	if (land) {
		*FireMask = CLASS_NO_FIRE;
		*ExtFireMask = ECLASS_CLEAR_LAND;
		}
	else {
		*FireMask = CLASS_WATER;
		*ExtFireMask = ECLASS_CLEAR_WATER;
		}


	if (day)
		fire = test1 || (test2 && test3 && test4 && (test5 || test6));
	else
		fire = test1 || (test2 && test3 && test4);


	if (!fire) {
		/* we have a non-fire pixel */

		if (land) {
			/* distinguish between "no fire" and "unknown" */
			if (nvalid > 0)
				counts->NumNoFirePix++;
			else {
				*FireMask = CLASS_UNKNOWN;
				*ExtFireMask = ECLASS_UNKNOWN_LAND;
				counts->NumUnknownLandPix++;
				}
			}
		else {
			/* distinguish between "no fire" and "unknown" */
			if (nvalid > 0)
				counts->NumNoFirePix++;
			else {
				*FireMask = CLASS_UNKNOWN;
				*ExtFireMask = ECLASS_UNKNOWN_WATER;
				counts->NumUnknownWaterPix++;
				}
			}

		return 0;
		}


	/* ----- we have a fire pixel ----- */

	if (land && nref) {
		compute_ndvi(BuffR5, BuffR7, nref, BuffNDVI);
		nwater_cont = bkg_water_pix(BuffR7, BuffR11,
			BuffNDVI, nref);
		compute_stats(BuffR7, nref, &MeanR7, &MadR7);
		}
	else {
		nwater_cont = 0;
		MeanR7 = MadR7 = -1.0F;
		}


	CountAdjacentPixels(i, j, PrevScan, scan, NextScan,
		&AdjCloud, &AdjWater);


	/* ----- rejection tests ----- */

	/* override detection if sun glint is present */
	nwc = nwater + ncoast;
	if ( (scan->pixflags[i][j].glint_level > 1) ||
		((AdjWater || nwc) && scan->pixflags[i][j].glint_level) ) {
		*QA |= 0x01000000;	/* bit 24 */
		counts->NumGlintRejectPix++;
		counts->NumNoFirePix++;
		return 0;
		}


	if (land) {
		/* ----- land-specific rejection tests ----- */

		/* override land detection if it's likely that this is a
		false alarm caused by excessive rejection of legitimate
		background pixels */
		if (day && !test1 && (fbfire > 0.1F) && (nbfire > 3) &&
			(BkgFireMeanT13 < 345.0F) &&
			(BkgFireMadT13 < 3.0F) && (scan->R7[i][j] > 0.15F) &&
			(T13 < (BkgFireMeanT13 + 6.0F*BkgFireMadT13))) {
			*QA |= 0x02000000;	/* bit 25 */
			counts->NumHotRejectPix++;
			counts->NumNoFirePix++;
			return 0;
			}

		/* override land fire pixels that are most likely to be
		false alarms caused by water-pixel contamination
		of the background */
		if (day && !test1 && nwater_cont) {
			*QA |= 0x04000000;	/* bit 26 */
			counts->NumCoastRejectLandPix++;
			counts->NumNoFirePix++;
			return 0;
			}

		/* Amazon forest-clearing rejection test */
		if (day && nref && (MeanR7 > 0.28F) && (T13 < 325.0F)
			&& (z15 > 3.7F)) {
			*QA |= 0x08000000;	/* bit 27 */
			counts->NumClearingRejectPix++;
			counts->NumNoFirePix++;
			return 0;
			}
		}
	else {
		/* ----- water-specific rejection tests ----- */

		/*
		Override water detection if it's likely that this is a
		false alarm caused by an unmasked land or coast pixel.
		The presence of any land or coast pixels in the background
		window is a good indication that this is a false alarm.
		*/
		if (!test1 && (nland || ncoast)) {
			*QA |= 0x04000000;	/* bit 26 */
			counts->NumCoastRejectWaterPix++;
			counts->NumNoFirePix++;
			return 0;
			}
		}


	C = confidence(T13, z13, DT, MeanDT, MadDT, AdjCloud,
		AdjWater * land,
		day, pfthresh->thresh_PF_T13[i][j]);

	*FireMask = CLASS_FIRE_LOW + (C >= 30) + (C >= 80);
	counts->NumFirePix++;
	if (land) {
		*ExtFireMask = ECLASS_FIRE_LAND;
		counts->NumLandFirePix++;
		}
	else {
		*ExtFireMask = ECLASS_FIRE_WATER;
		counts->NumWaterFirePix++;
		}

	if (nvalid > 0) {
		/* compute mean L13 */
		compute_stats(BuffL13, nvalid, &MeanL13, &fdummy);

		power = fire_power((float) scan->L13[i][j], MeanL13, j);
		}
	else
		power = 0.0F;


	cmg_row_col((float) scan->lat[i][j], (float) scan->lon[i][j],
		&cmg_row, &cmg_col);

	RelAzAng = fixang(scan->SolAzAng[i][j] - scan->ViewAzAng[i][j]);

	st = append_fire_pixel(i + first_line, j,
		scan->lat[i][j], scan->lon[i][j],
		scan->R7[i][j], T13, T15, MeanT13, MeanT15, MeanDT,
		MadT13, MadT15, MadDT, AdjCloud, AdjWater,
		winparam, nvalid, C, power, land,
		MeanR7, MadR7, scan->ViewZenAng[i][j], scan->SolZenAng[i][j],
		RelAzAng, cmg_row, cmg_col);
	if (st) return st;

	if (AdjCloud) {
		*QA |= 0x100000;	/* bit 20 (was bit 21 in C5) */
		counts->NumCloudAdjFirePix++;
		}

/* FIX: may not be necessary to include adjacent water pixels for water fire pixels */
	if (AdjWater) {
		*QA |= 0x200000;	/* bit 21 (was bit 22 in C5) */
		counts->NumWaterAdjFirePix++;
		}

	L = 2*winparam + 1;

	return 0;
}






static int BkgPixels(int ic, int jc, int surf,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT13, float *BuffT15, float *BuffDT, float *BuffL13,
	float *BuffR5, float *BuffR7, float *BuffR11, int *nref,
	float *BkgFireBuffT13, float *BkgFireBuffT15,
	float *BkgFireBuffDT, int *nbfire,
	int *winparam, int *nwater, int *nland, int *ncoast)
/*
!C*****************************************************************************
!Description:
	Fill buffers with M13 and M15 brightness temperatures (T13 and T15),
	and their difference (DT), for valid background pixels contained in the
	smallest square background window needed to characterize the background
	of a potential fire pixel.

	Window starts as 3 x 3 pixels in size, always excluding the two
	along-scan pixels adjacent to the candidate fire pixel, and is expanded
	as necessary until either a sufficient number of background pixels have
	been acquired, or the window becomes 21 x 21 pixels in size a
	sufficient number of background pixels still has not been obtained.  In
	this case, the characterization is deemed unsuccessful.

!Input Parameters:
	ic		scan position of potential fire pixel
	jc		sample position of potential fire pixel
	PrevScan
	CurrScan
	NextScan

!Output Parameters:
	BuffT13		buffer containing T13 for valid background pixels
	BuffT15		buffer containing T15 for valid background pixels
	BuffDT		buffer containing DT for valid background pixels
BuffL13		buffer containing L13 ""

	BuffR5
	BuffR7
	BuffR11
	
	nref

	BkgFireBuffT13
	BkgFireBuffT15
	BkgFireBuffDT

	nbfire

	winparam	size parameter for window needed to characterize
			background

	nwater
	nland
	ncoast

!Input/Output Parameters:

!Returns:
	Number of valid background pixels, or 0 if an insufficient number of
	valid background pixels could be obtained.

!Revision History
	Original version 1997, L. Giglio.
	May 2002 updated code to reflect version 4 algorithm changes.
	Dec 2008 code clean-up.

!Team-unique Header:

!References and Credits

!Design Notes
	Window size is related to size parameter as:

		window size in pixels = 2 * winparam + 1

!END***************************************************************************
*/
{
	int nvalid, minvalid;
	int winsize;
	int r;
	const int rmax = 10;	/* corresponds to 21 by 21 window */


	nvalid = *nref = *nbfire = *nwater = *nland = *ncoast = 0;
	r = 1;		/* start at 3 x 3 ring */
	while (r <= rmax) {
		*winparam = r;
		winsize = 2*r + 1;

		GetValidBkgPixels(ic, jc, r, surf, PrevScan, CurrScan, NextScan,
			BuffT13, BuffT15, BuffDT, BuffL13, &nvalid,
			BuffR5, BuffR7, BuffR11, nref,
			BkgFireBuffT13, BkgFireBuffT15, BkgFireBuffDT, nbfire,
			nwater, nland, ncoast);

		minvalid = (int) (0.25F * (winsize*winsize - 3));		/* 25% */
		if (minvalid < 8) minvalid = 8;

		if (nvalid >= minvalid) return nvalid;

		r++;
		}

	*winparam = *nref = *nbfire = *nwater = *nland = *ncoast = 0;
	return 0;
}



static void GetValidBkgPixels(int ic, int jc, int r, int surf_state,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	float *BuffT13, float *BuffT15, float *BuffDT, float *BuffL13,
	int *nvalid,
	float *BuffR5, float *BuffR7, float *BuffR11, int *nref,
	float *BkgFireBuffT13, float *BkgFireBuffT15, float *BkgFireBuffDT,
	int *nbfire, int *nwater, int *nland, int *ncoast)
/*
!C*****************************************************************************
!Description:
	Append ch. 13, ch. 13, and ch. 13 - ch. 15 brightness temperatures of
	valid background pixels from a square "ring" of pixels to buffers.

!Input Parameters:
	ic		scan position around which ring will be centered
	jc		sample position around which ring will be centered
	r		"radius" of square ring

	surf_state	surface type for which we are to accumulate pixels
			(land or water)

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	BuffT13		buffer containing T13 (K) for valid background pixels
	BuffT15		buffer containing T15 (K) for valid background pixels
	BuffDT		buffer containing DT (K) for valid background pixels
	BuffL13		buffer containing L13 (W/m^2/str/micron) "

!Input/Output Parameters:
	nvalid		number of valid background pixels contained in buffers
	nref
	nbfire
	nwater		number of water pixels present in background window
	nland		number of land pixels present in background window
	ncoast		number of coastal pixels present in background window

!Returns:

!Revision History
	Original version 1997, L. Giglio.

	Modified to deal with MODIS scan bow tie overlap near scan edges,
	allowing contextual algorithm to be applied even way off nadir.
	L. Giglio, Feb. 2002.

	Modified to also tabulate values for all background fire pixels,
	Mar. 2002.

	Oct. 2010, L. Giglio.  Heavily modified to accomodate processing of
	water pixels.

!END***************************************************************************
*/
{
	int i, j;
	int jstart, jstop;
	int jl, jr;		/* left and right samples */
	Scan *scan;

	int icorr;		/* bow-tie-corrected line index */


	if ( (jstart = jc - r) < 0) jstart = 0;
	if ( (jstop = jc + r) >= SAMPLES_750M) jstop = SAMPLES_750M - 1;



	/* top edge */
	i = ic - r;
	if (i < 0) {
		/* look in previous scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i + LINES_PER_SCAN_750M;
			if (icorr < 0) continue;

			if (PrevScan->pixflags[icorr][j].surf == LAND)
				(*nland)++;
			else {
				if (PrevScan->pixflags[icorr][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (PrevScan->pixflags[icorr][j].surf != surf_state)
				continue;

			if (PrevScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT13[*nvalid] = PrevScan->T13[icorr][j];
				BuffT15[*nvalid] = PrevScan->T15[icorr][j];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = PrevScan->L13[icorr][j];
				(*nvalid)++;

				if (PrevScan->R7[icorr][j] > 0.0F) {
					BuffR5[*nref] = PrevScan->R5[icorr][j];
					BuffR7[*nref] = PrevScan->R7[icorr][j];
					BuffR11[*nref] = PrevScan->R11[icorr][j];
					(*nref)++;
					}
				}

			if (PrevScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT13[*nbfire] =
					PrevScan->T13[icorr][j];
				BkgFireBuffT15[*nbfire] =
					PrevScan->T15[icorr][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}
	else {
		/* no bowtie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf == LAND)
				(*nland)++;
			else {
				if (CurrScan->pixflags[i][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (CurrScan->pixflags[i][j].surf != surf_state)
				continue;

			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT13[*nvalid] = CurrScan->T13[i][j];
				BuffT15[*nvalid] = CurrScan->T15[i][j];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = CurrScan->L13[i][j];
				(*nvalid)++;

				if (CurrScan->R7[i][j] > 0.0F) {
					BuffR5[*nref] = CurrScan->R5[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					BuffR11[*nref] = CurrScan->R11[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT13[*nbfire] = CurrScan->T13[i][j];
				BkgFireBuffT15[*nbfire] = CurrScan->T15[i][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}



	/* bottom edge */
	i = ic + r;
	if (i >= LINES_PER_SCAN_750M) {
		/* look in next scan, handling bow tie */

		for (j = jstart; j <= jstop; j++) {
			icorr = i - LINES_PER_SCAN_750M;
			if (icorr >= LINES_PER_SCAN_750M) continue;

			if (NextScan->pixflags[icorr][j].surf == LAND)
				(*nland)++;
			else {
				if (NextScan->pixflags[icorr][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (NextScan->pixflags[icorr][j].surf != surf_state)
				continue;

			if (NextScan->pixflags[icorr][j].valid_bkg_pix) {
				BuffT13[*nvalid] = NextScan->T13[icorr][j];
				BuffT15[*nvalid] = NextScan->T15[icorr][j];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = NextScan->L13[icorr][j];
				(*nvalid)++;

				if (NextScan->R7[icorr][j] > 0.0F) {
					BuffR5[*nref] = NextScan->R5[icorr][j];
					BuffR7[*nref] = NextScan->R7[icorr][j];
					BuffR11[*nref] = NextScan->R11[icorr][j];
					(*nref)++;
					}
				}

			if (NextScan->pixflags[icorr][j].bkg_fire) {
				BkgFireBuffT13[*nbfire] =
					NextScan->T13[icorr][j];
				BkgFireBuffT15[*nbfire] =
					NextScan->T15[icorr][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}
	else {
		/* no bowtie */
		for (j = jstart; j <= jstop; j++) {
			if (CurrScan->pixflags[i][j].surf == LAND)
				(*nland)++;
			else {
				if (CurrScan->pixflags[i][j].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (CurrScan->pixflags[i][j].surf != surf_state)
				continue;

			if (CurrScan->pixflags[i][j].valid_bkg_pix) {
				BuffT13[*nvalid] = CurrScan->T13[i][j];
				BuffT15[*nvalid] = CurrScan->T15[i][j];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = CurrScan->L13[i][j];
				(*nvalid)++;

				if (CurrScan->R7[i][j] > 0.0F) {
					BuffR5[*nref] = CurrScan->R5[i][j];
					BuffR7[*nref] = CurrScan->R7[i][j];
					BuffR11[*nref] = CurrScan->R11[i][j];
					(*nref)++;
					}
				}

			if (CurrScan->pixflags[i][j].bkg_fire) {
				BkgFireBuffT13[*nbfire] = CurrScan->T13[i][j];
				BkgFireBuffT15[*nbfire] = CurrScan->T15[i][j];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}



	/* if doing 3 x 3 window, i.e. r == 1, skip pixels adjacent to the
	candidate fire pixel in the along-scan direction */
	if (r == 1) return;



	jl = jc - r;	/* left edge */
	jr =  jc + r;	/* right edge */

	/*
	value of "i"			desired value of "scan"
	------------			-----------------------
	i < 0				scan = PrevPtr

	0 <= i < LINES_PER_SCAN_750M	scan = CurrPtr

	i >= LINES_PER_SCAN_750M		scan = NextPtr
	*/

	/* left side */
	if (jl >= 0) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + LINES_PER_SCAN_750M;
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_750M) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_750M;
				if (icorr >= LINES_PER_SCAN_750M) continue;
				}

			if (scan->pixflags[icorr][jl].surf == LAND)
				(*nland)++;
			else {
				if (scan->pixflags[icorr][jl].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (scan->pixflags[icorr][jl].surf != surf_state)
				continue;

			if (scan->pixflags[icorr][jl].valid_bkg_pix) {
				BuffT13[*nvalid] = scan->T13[icorr][jl];
				BuffT15[*nvalid] = scan->T15[icorr][jl];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = scan->L13[icorr][jl];
				(*nvalid)++;

				if (scan->R5[icorr][jl] > 0.0F) {
					BuffR5[*nref] = scan->R5[icorr][jl];
					BuffR7[*nref] = scan->R7[icorr][jl];
					BuffR11[*nref] = scan->R11[icorr][jl];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jl].bkg_fire) {
				BkgFireBuffT13[*nbfire] = scan->T13[icorr][jl];
				BkgFireBuffT15[*nbfire] = scan->T15[icorr][jl];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}


	/* right side */
	if (jr < SAMPLES_750M) {
		for (i = ic - r + 1; i < ic + r; i++) {
			/* initialize as if pixel is in current scan buffer */
			scan = CurrScan;
			icorr = i;

			/* ...but reset if necessary */
			if (i < 0) {
				scan = PrevScan;
				icorr = i + LINES_PER_SCAN_750M;
				if (icorr < 0) continue;
				}
			else if (i >= LINES_PER_SCAN_750M) {
				scan = NextScan;
				icorr = i - LINES_PER_SCAN_750M;
				if (icorr >= LINES_PER_SCAN_750M) continue;
				}

			if (scan->pixflags[icorr][jr].surf == LAND)
				(*nland)++;
			else {
				if (scan->pixflags[icorr][jr].surf == COAST)
					(*ncoast)++;
				else
					(*nwater)++;
				}

			if (scan->pixflags[icorr][jr].surf != surf_state)
				continue;

			if (scan->pixflags[icorr][jr].valid_bkg_pix) {
				BuffT13[*nvalid] = scan->T13[icorr][jr];
				BuffT15[*nvalid] = scan->T15[icorr][jr];
				BuffDT[*nvalid] = BuffT13[*nvalid] -
					BuffT15[*nvalid];
				BuffL13[*nvalid] = scan->L13[icorr][jr];
				(*nvalid)++;

				if (scan->R7[icorr][jr] > 0.0F) {
					BuffR5[*nref] = scan->R5[icorr][jr];
					BuffR7[*nref] = scan->R7[icorr][jr];
					BuffR11[*nref] = scan->R11[icorr][jr];
					(*nref)++;
					}
				}

			if (scan->pixflags[icorr][jr].bkg_fire) {
				BkgFireBuffT13[*nbfire] = scan->T13[icorr][jr];
				BkgFireBuffT15[*nbfire] = scan->T15[icorr][jr];
				BkgFireBuffDT[*nbfire] =
					BkgFireBuffT13[*nbfire] -
					BkgFireBuffT15[*nbfire];
				(*nbfire)++;
				}
			}
		}
}



static void CountAdjacentPixels(int ic, int jc,
	Scan *PrevScan, Scan *CurrScan, Scan *NextScan,
	int *AdjCloud, int *AdjWater)
/*
!C*****************************************************************************
!Description:
	Count number of cloud and water pixels adjacent to detected
	fire pixel.  Here both coast and water pixels are considered to be
	water.

!Input Parameters:
	ic
	jc

	PrevScan,
	CurrScan,
	NextScan	pointers to previous, current, and next scan buffers

!Output Parameters:
	AdjCloud
	AdjWater

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Ported from Collection 6 MODIS code.

!References and Credits

!Design Notes
	Adjacent pixels in the trimmed region at the edge of the VIIRS scan
	are not counted properly for the first and last row of the scan,
	resulting in a possible undercount for pixels in either of those
	rows.  This will be fixed in a subsequent version.

!END***************************************************************************
*/
{
	int i1, i2, j;
	Scan *scan1, *scan2;
	int jstart, jstop;

	*AdjCloud = *AdjWater = 0;


	/* pixels to the left and right of center pixel */
	if (jc > 0) {
		if (CurrScan->pixflags[ic][jc-1].surf != LAND) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc-1].cloud) (*AdjCloud)++;
		}

	if (jc < (SAMPLES_750M - 1)) {
		if (CurrScan->pixflags[ic][jc+1].surf != LAND) (*AdjWater)++;
		if (CurrScan->pixflags[ic][jc+1].cloud) (*AdjCloud)++;
		}


	/* sets of three pixels above and below center pixel */
	if ( (jstart = jc - 1) < 0) jstart = 0;
	if ( (jstop = jc + 1) >= SAMPLES_750M) jstop = SAMPLES_750M - 1;

	switch (ic) {
		case 0:
			i1 = LINES_PER_SCAN_750M - 1;
			i2 = 1;
			scan1 = PrevScan;
			scan2 = CurrScan;
			break;

		case (LINES_PER_SCAN_750M - 1):
			i1 = ic - 1;
			i2 = 0;
			scan1 = CurrScan;
			scan2 = NextScan;
			break;

		default:
			i1 = ic - 1;
			i2 = ic + 1;
			scan1 = scan2 = CurrScan;
			break;
		}


	for (j = jstart; j <= jstop; j++) {
		if (scan1->pixflags[i1][j].surf != LAND) (*AdjWater)++;
		if (scan1->pixflags[i1][j].cloud) (*AdjCloud)++;

		if (scan2->pixflags[i2][j].surf != LAND) (*AdjWater)++;
		if (scan2->pixflags[i2][j].cloud) (*AdjCloud)++;
		}
}




/*
Alg. will only call this function with n >= 1 so no checks for n == 0
are needed.
*/
static void compute_stats(float *x, int n, float *mean, float *mad)
{
	int	i;
	double	sum;

	/* mean */
	for (i = 0, sum = 0.0; i < n; i++) sum += (double) x[i];
	*mean = (float) (sum / n);

	/* mean absolute deviation */
	for (i = 0, sum = 0.0; i < n; i++)
		sum += (double) fabsf(x[i] - *mean);

	*mad = (float) (sum / n);
}






static int SunGlint(float R5, float R7, float R11, float ts, float tv,
	float phi, float *GlintAng)
/*
!C*****************************************************************************
!Description:
	Perform sun-glint test specified in Giglio et al. (2003), RSE.

!Input Parameters:
	R5	M5 (0.64 micron) 750 m reflectance
	R7	M7 (0.86 micron) 750 m reflectance
	R11	M11 (2.1 micron)  750 m reflectance
	ts	solar zenith angle ("theta s") in degrees
	tv	view zenith angle ("theta v") in degrees
	phi	relative azimuth angle in degrees

!Output Parameters:
	GlintAng	"glint" angle (degrees)

!Input/Output Parameters:
	None.

!Returns:
	0 if sun glint absent, otherwise glint level from 1 (lowest)
	to 3 (highest)

!Revision History
	Adapted from MODIS Collection 6 fire code, L. Giglio, 2012.

!END***************************************************************************
*/
{
	float cos_tr;		/* cosine of reflected sun angle "theta r" */
	float tr;

	cos_tr = cosf(D2RF * tv) * cosf(D2RF * ts) -
		sinf(D2RF * tv) * sinf(D2RF * ts) * cosf(D2RF * phi);

	tr = R2D * acosf(cos_tr);

	*GlintAng = tr;

	/* automatically assume glint for very small glint angles */
	if (tr < 2.0F) return 3;

        if ((tr < 10.0F) && (R5 > 0.10F) && (R7 > 0.20F) && (R11 > 0.12F))
		return 2;

	if (tr < 15.0F) return 1;

	return 0;
}




static float fixang(float ang)
/*
!C*****************************************************************************
!Description:
	Map angle in range [-540,+540] to range [-180,+180].

!Input Parameters:
	ang	angle (degrees)

!Output Parameters:
	None.

!Input/Output Parameters:
	None.

!Returns:
	Returns angle mapped to range [-180,+180]

!Revision History

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	if (ang > +180.0F) return ang - 360.0F;
	if (ang <= -180.0F) return 360.0F + ang;
	return ang;
}






/* does NOT handle missing data in input arrays; make sure there isn't any when calling */
static void compute_ndvi(float *R5, float *R7, int n, float *NDVI)
{
	int i;

	for (i = 0; i < n; i++)
		NDVI[i] = (R7[i] - R5[i]) / (R5[i] + R7[i]);
}



/* returns number of background pixels that are very likely to contain water */
static int bkg_water_pix(float *R7, float *R11, float *NDVI, int n)
{
	int i;
	int nwater;

	for (i = nwater = 0; i < n; i++)
		if ((R7[i] < 0.15F) && (R11[i] < 0.05F) && (NDVI[i] < 0.0F))
			nwater++;

	return nwater;
}


