/*
!C******************************************************************

!Description: 
	Functions for dealing with the "fire pixel table" written into
	the VIIRS Level 2 fire product.  The fire pixel table is
	actually a collection of many one-dimensional SDS's containing
	information about each fire pixel identified within a granule.

	Originally the fire pixel table was supposed to be a single HDF
	vdata containing multiple fields, an HDF construct intended for
	storing tabular data.  However, because of a horrible limitation
	imposed upon vdatas by the HDF library (specifically, the sequence
	for opening and closing HDF files is mutually exclusive for SDS's
	and vdatas), and limitations within vdatas themselves, multiple SDS's
	had to be used instead.


!Revision History:
	Original version, August 1999, L. Giglio

        Revised 30-NOV-99: writes all data to a buffer instead of directly to
	the SDS (J. Ray).

	Completely rewritten Feb. 2001, L. Giglio.

	Rewritten May 2012 for VIIRS, L. Giglio.


!Design Notes
	The proper sequence of public functions within this module is:

		1. Call new_fire_pixel_table().

		2. Call append_fire_pixel() for each fire pixel
		identified within a granule.

		3. Call write_fire_pixel_table() after each granule has
		been processed.

	This should be done for each L2 fire product granule generated.

		4. Call free_fire_pixel_table() after all granules have been
		processed.

!END*************************************************************************/


#include "mfhdf.h"
#include "firetab.h"
#include "error.h"


extern int verbose;



/* structure to contain info. about each SDS comprising fire pixel table */
typedef struct {
	char	*name;
	int32	type;		/* HDF data type */
	char	*desc;		/* description (long name) */
	char	*unit;
	void	*data;
	} SDS;



enum {FPTAB_LINE, FPTAB_SAMPLE, FPTAB_LAT, FPTAB_LON, FPTAB_R7,
	FPTAB_T13, FPTAB_T15,
	FPTAB_MEAN13, FPTAB_MEAN15, FPTAB_MEANDT,
	FPTAB_SDEV13, FPTAB_SDEV15, FPTAB_SDEVDT,
	FPTAB_POWER, FPTAB_ADJCLOUD, FPTAB_ADJWATER,
	FPTAB_WINSIZE, FPTAB_NVALID, FPTAB_CONF, FPTAB_LAND,
	FPTAB_MEANR7, FPTAB_MADR7,
	FPTAB_VIEWZENANG, FPTAB_SOLZENANG, FPTAB_RELAZANG,
	FPTAB_CMG_ROW, FPTAB_CMG_COL,
	NUM_SDS};

/*
These are the separate SDS's comprising the "fire pixel table".
Order of table entries must match that of above enumeration.
*/
static SDS sds[NUM_SDS] = {
	{"FP_line",	DFNT_INT16,	"granule line of fire pixel",	(char *) NULL,	(void *) NULL},
	{"FP_sample",	DFNT_INT16,	"granule sample of fire pixel",	(char *) NULL,	(void *) NULL},
	{"FP_latitude",	DFNT_FLOAT32,	"latitude of fire pixel",	"degrees",	(void *) NULL},
	{"FP_longitude",	DFNT_FLOAT32,	"longitude of fire pixel",	"degrees",	(void *) NULL},

	{"FP_R7",		DFNT_FLOAT32,	"M7 reflectance "
					"of fire pixel",		(char *) NULL,	(void *) NULL},

	{"FP_T13",		DFNT_FLOAT32,	"M13 brightness "
					"temperature of fire pixel",	"kelvins",	(void *) NULL},

	{"FP_T15",		DFNT_FLOAT32,	"M15 brightness "
					"temperature of fire pixel",	"kelvins",	(void *) NULL},


 	{"FP_MeanT13",	DFNT_FLOAT32,	"M13 brightness "
					"temperature of background",	"kelvins", (void *) NULL},

	{"FP_MeanT15",	DFNT_FLOAT32,	"M15 brightness "
					"temperature of background",	"kelvins", (void *) NULL},

	{"FP_MeanDT",	DFNT_FLOAT32,	"mean background brightness "
					"temperature difference",	"kelvins", (void *) NULL},

	{"FP_MAD_T13",	DFNT_FLOAT32,	"background M13 brightness "
					"temperature mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_MAD_T15",	DFNT_FLOAT32,	"background M15 brightness "
					"temperature mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_MAD_DT",	DFNT_FLOAT32,	"background brightness temperature "
					"difference mean absolute deviation", "kelvins", (void *) NULL},

	{"FP_power",	DFNT_FLOAT32,	"fire radiative power",			"MW", (void *) NULL},
	{"FP_AdjCloud",	DFNT_UINT8,	"number of adjacent cloud pixels",	(char *) NULL, (void *) NULL},
	{"FP_AdjWater",	DFNT_UINT8,	"number of adjacent water pixels",	(char *) NULL, (void *) NULL},
	{"FP_WinSize",	DFNT_UINT8,	"background window size",		(char *) NULL, (void *) NULL},
	{"FP_NumValid",	DFNT_INT16,	"number of valid background pixels",	(char *) NULL, (void *) NULL},
	{"FP_confidence", DFNT_UINT8,	"detection confidence",			"%", (void *) NULL},
	{"FP_land", DFNT_UINT8,		"land pixel flag",			(void *) NULL,	(void *) NULL},
	{"FP_MeanR7",	DFNT_FLOAT32,	"background M7 reflectance",		(char *) NULL,	(void *) NULL},
	{"FP_MAD_R7",	DFNT_FLOAT32,	"background M7 reflectance mean absolute deviation", (char *) NULL, (void *) NULL},
	{"FP_ViewZenAng", DFNT_FLOAT32,	"view zenith angle", "degrees",	(void *) NULL},
	{"FP_SolZenAng", DFNT_FLOAT32,	"solar zenith angle", "degrees",	(void *) NULL},
	{"FP_RelAzAng",	DFNT_FLOAT32,	"relative azimuth angle", "degrees",	(void *) NULL},
	{"FP_CMG_row",	DFNT_INT16,	"CMG row",	(char *) NULL,	(void *) NULL},
	{"FP_CMG_col",	DFNT_INT16,	"CMG column",	(char *) NULL,	(void *) NULL}
	};





static long nfire = 0L;
static long nalloc = 0L;


#define ALLOC_ELEMS 1000


int append_fire_pixel(int line, int sample, float lat, float lon,
	float R7, float T13, float T15, float MeanT13, float MeanT15,
	float MeanDT, float MAD_T13, float MAD_T15, float MAD_DT,
	int AdjCloud, int AdjWater, int winparam, int nvalid, int confidence,
	float power, int land, float MeanR7, float MadR7,
	float ViewZenAng, float SolZenAng, float RelAzAng,
	int cmg_row, int cmg_col)
{
	int i;
	uint8	*bptr;
	int16	*iptr;
	float32	*fptr;
	size_t bytes;

	if (nfire >= nalloc) {
		nalloc += ALLOC_ELEMS;

#ifdef DEBUG
	if (verbose) printf("FPTAB: reallocating %ld elements.\n", nalloc);
#endif

		for (i = 0; i < NUM_SDS; i++) {
			bytes = DFKNTsize(sds[i].type) * nalloc;

			if (!(sds[i].data = realloc(sds[i].data, bytes))) return error(ERR_F_MEM);
			}
		}


	iptr = (int16 *) sds[FPTAB_LINE].data;
	iptr[nfire] = (int16) line;

	iptr = (int16 *) sds[FPTAB_SAMPLE].data;
	iptr[nfire] = (int16) sample;

	fptr = (float32 *) sds[FPTAB_LAT].data;
	fptr[nfire] = (float32) lat;

	fptr = (float32 *) sds[FPTAB_LON].data;
	fptr[nfire] = (float32) lon;

	fptr = (float32 *) sds[FPTAB_R7].data;
	fptr[nfire] = (float32) R7;

	fptr = (float32 *) sds[FPTAB_T13].data;
	fptr[nfire] = (float32) T13;

	fptr = (float32 *) sds[FPTAB_T15].data;
	fptr[nfire] = (float32) T15;

	fptr = (float32 *) sds[FPTAB_MEAN13].data;
	fptr[nfire] = (float32) MeanT13;

	fptr = (float32 *) sds[FPTAB_MEAN15].data;
	fptr[nfire] = (float32) MeanT15;

	fptr = (float32 *) sds[FPTAB_MEANDT].data;
	fptr[nfire] = (float32) MeanDT;

	fptr = (float32 *) sds[FPTAB_SDEV13].data;
	fptr[nfire] = (float32) MAD_T13;

	fptr = (float32 *) sds[FPTAB_SDEV15].data;
	fptr[nfire] = (float32) MAD_T15;

	fptr = (float32 *) sds[FPTAB_SDEVDT].data;
	fptr[nfire] = (float32) MAD_DT;

	fptr = (float32 *) sds[FPTAB_POWER].data;
	fptr[nfire] = (float32) power;

	bptr = (uint8 *) sds[FPTAB_ADJCLOUD].data;
	bptr[nfire] = (uint8) AdjCloud;

	bptr = (uint8 *) sds[FPTAB_ADJWATER].data;
	bptr[nfire] = (uint8) AdjWater;

	iptr = (int16 *) sds[FPTAB_NVALID].data;
	iptr[nfire] = (int16) nvalid;

	bptr = (uint8 *) sds[FPTAB_WINSIZE].data;
	if (winparam)
		bptr[nfire] = (uint8) (2 * winparam + 1);
	else
		bptr[nfire] = (uint8) 0;

	bptr = (uint8 *) sds[FPTAB_CONF].data;
	bptr[nfire] = (uint8) confidence;

	bptr = (uint8 *) sds[FPTAB_LAND].data;
	bptr[nfire] = (uint8) (land != 0);


	fptr = (float32 *) sds[FPTAB_MEANR7].data;
	fptr[nfire] = (float32) MeanR7;

	fptr = (float32 *) sds[FPTAB_MADR7].data;
	fptr[nfire] = (float32) MadR7;


	fptr = (float32 *) sds[FPTAB_VIEWZENANG].data;
	fptr[nfire] = (float32) ViewZenAng;

	fptr = (float32 *) sds[FPTAB_SOLZENANG].data;
	fptr[nfire] = (float32) SolZenAng;

	fptr = (float32 *) sds[FPTAB_RELAZANG].data;
	fptr[nfire] = (float32) RelAzAng;


	iptr = (int16 *) sds[FPTAB_CMG_ROW].data;
	iptr[nfire] = (int16) cmg_row;

	iptr = (int16 *) sds[FPTAB_CMG_COL].data;
	iptr[nfire] = (int16) cmg_col;

	nfire++;

	return 0;
}



void free_fire_pixel_table(void)
{
	int i;

	for (i = 0; i < NUM_SDS; i++) if (sds[i].data) free(sds[i].data);

	nfire = nalloc = 0L;
}



void new_fire_pixel_table(void)
{
	nfire = 0L;
}




int write_fire_pixel_table(int32 sd_id, int write_cmg_fields)
/*
!C*****************************************************************************

!Description:
	Create and writes the SDS's comprising the fire pixel table
	which has been stored internally up to this point.

!Input Parameters:
	sd_id  HDF SD ID.

	Uses private variables "nfire" and array "sds".

!Output Parameters:
	None.

!Input/Ouput Parameters:
	None.

!Returns:
	Returns 0 if successful, non-zero otherwise.

!Revision History:
	Original version, L. Giglio (August 1999)
	Modified Feb. 2001 (L. Giglio) for V3 fire code.

!Team-unique Header:
	See any of the other fifty dozen functions in this code for this
	information.

!References and Credits:
	See any of the other fifty dozen functions in this code for this
	information.

!Design Notes


!END
****************************************************************************/
{
	int i;
	int32 sds_id;
	int32 dim_sizes[1];
	int32 start[1], edges[1];
	int32 dim_id;

	int skip_cmg_fields;


	skip_cmg_fields = !write_cmg_fields;


	dim_sizes[0] = edges[0] = nfire;
	start[0] = 0;

	for (i = 0; i < NUM_SDS; i++) {
		if (skip_cmg_fields && ((i == FPTAB_CMG_ROW) ||
			(i == FPTAB_CMG_COL))) continue;

		if ((sds_id = SDcreate(sd_id, sds[i].name, sds[i].type, 1, dim_sizes)) == FAIL)
			return error(ERR_HDF);
		if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL)
			return error(ERR_HDF);
		if (SDsetdimname(dim_id, "number_of_active_fires") == FAIL)
			return error(ERR_HDF);
		if (SDsetdatastrs(sds_id, sds[i].desc, sds[i].unit, NULL, NULL) == FAIL)
			return error(ERR_HDF);

		/* only write SDS if we have at least one fire pixel */
		if (nfire)
			if (SDwritedata(sds_id, start, NULL, edges, sds[i].data) == FAIL)
				return error(ERR_HDF);

		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}
 
	return 0;
}

