#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "const.h"
#include "frp.h"


static double VIIRS_scan_angle(int sample, int *numagg);
static float VIIRS_pixel_area(int sample);



float fire_power(float L13, float MeanL13, int sample)
/*
!C*****************************************************************************
!Description:
	Implementation of Wooster's fire radiative power calculation for
	the VIIRS sensor.

!Input Parameters:
	L13
	MeanL13
	sample (0 - 3199, inclusive)

!Output Parameters:
	Returns fire power in MW.

!Revision History:
	Original version March 2013, L. Giglio.
	Based on C6 MODIS fire code.


!END***************************************************************************
*/
{
	double FRP;
	double Apix;			/* square meters */
	const double sigma = 5.6704e-8;	/* [W m^-2 K^-4] Stefan-Boltzman constant */

	/* empirical constant for VIIRS M13 @ 4.05 micron central wavelength */
	const double a = 2.88e-9;


	/* unable to perform retrieval without background radiance */
	if (MeanL13 <= 0.0F) return 0.0F;

	Apix = VIIRS_pixel_area(sample) * 1.0e6;	/* m^2 */

	FRP = (Apix * sigma / a) * (L13 - MeanL13) * 1.0e-6;

/*printf("%9.1f m^2  %10.1f\n", Apix, FRP);*/
	return (float) FRP;
}




/*
Calculate approximate area of aggregated VIIRS moderate resolution pixel in
km^2 at specified position within scan.

0 <= sample <= 3199

*/
static float VIIRS_pixel_area(int sample)
{
	const double h = 824.0;		/* NPP orbit altitude [km] */
	const double pt = 0.721;	/* along-track nadir pixel size [km] */
	double ps;			/* along-scan nadir pixel size [km] */
	double r, ss, st;
	double theta, cos_theta;
	double temp, sqrt_temp;
	double DS, DT, area;

	int agg_pix;			/* 1, 2, or 3 */


	theta = VIIRS_scan_angle(sample, &agg_pix) * D2R;
	cos_theta = cos(theta);


	ps = 0.256 * agg_pix;		/* [km] */

	r = EARTH_RADIUS + h;		/* [km] */

	ss = ps / h;		/* [rad] */
	st = pt / h;		/* [rad] */


	temp = pow(EARTH_RADIUS/r, 2.0) - pow(sin(theta), 2.0);
	sqrt_temp = sqrt(temp);

	DS = EARTH_RADIUS * ss * (cos_theta/sqrt_temp - 1.0);
	DT = r * st * (cos_theta - sqrt_temp);
	area = DS * DT;

	return area;
}



/*
Compute scan angle (in degrees) for given VIIRS M-band sample.

	0 <= sample <= 3199
*/
static double VIIRS_scan_angle(int sample, int *numagg)
{
	int rel_sample;
	int k;
	double scan_angle;

	const double scan_step = 56.06 / 3152;	/* degrees */

	/* lookup tables indexed by aggregation zone (variable "k" in code
 	below) */
	/* number of pixels aggregated in each zone */
	static int aggpix[] = {3,2,1};
	static int suboff[] = {0, 592, 960};
	static int addoff[] = {0, 1776, 2512};


	rel_sample = abs(sample - 1600);


	/*
	compute index into aggregation zone (0, 1, or 2)
		0 is center zone (three-pixel aggregation)
		1 is middle zone (two-pixel aggregation)
		2 is edge zone (one-pixel aggregation, i.e., no aggregation)
	*/
	k = 2 - ((rel_sample <= 592) + (rel_sample <= 960));

	*numagg = aggpix[k];

	scan_angle = (addoff[k] + (rel_sample - suboff[k]) *
		aggpix[k]) * scan_step;
/*printf("%4d, ang %6.2f, nagg %3d\n", sample, scan_angle, *numagg);*/

	return scan_angle;
}

