/*
!C******************************************************************
!Description:
	Functions to handle input from HDF4 files.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio and B. Wind, March 2013
	W. Walsh, May 2014

!Design Notes:

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
/*#include "io.h"*/
#include "input.h"
#include "error.h"
#include "misc.h"


#define OPTIONAL_SCALE_OFFSET 1


static int identify_satellite(int32 sd_id_vmae, enum Satellite *sat);
static char *DimStr(int32 rank, int32 *dim_sizes);

static int init_input_sds(int32 sd_id, char *sds_name, int32 *sds_index_ptr,
	int32 expected_data_type, int32 dim_sizes[MAX_VAR_DIMS],
	float32 *scale, float32 *offset, int optional_scale);

static int read_scale_offset(int32 sds_id, float32 *scale, float32 *offset,
	int optional);

static void unscale(uint16 ScaledVal[LINES_PER_SCAN_750M][SAMPLES_750M],
	float32 UnscaledVal[LINES_PER_SCAN_750M][SAMPLES_750M],
	float scale, float offset);

static int read_band_scan(int32 sd_id, int32 sds_index, int32 start[2],
	int32 edges[2], void *buff);


/* from ADL source file "ProCmnDefs.h" */
static const uint16 UINT16_FILL_TEST = 65527;


static const char *GRINGLATATTNAM1="Unagg_GRingLatitude";
static const char *GRINGLONATTNAM1="Unagg_GRingLongitude";


extern int verbose;



/*
	sd_id_vmae	NPP_VMAE SDR (L1B) product
*/
int InitInputHDF(int32 sd_id_vmae, int32 sd_id_lc, HDF_InputInfo *hinfo)
{
	int32 dim_sizes[MAX_VAR_DIMS];
	int st;


	hinfo->sd_id_vmae = sd_id_vmae;
	hinfo->sd_id_lc = sd_id_lc;


	/* ---------- VMAE ---------- */

	/* M13 brightness temperature */
	if ( (st = init_input_sds(sd_id_vmae, "BrightnessTemperature_M13",
		&hinfo->sds_index_M13, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* Number of image lines should be a multiple of 16 to match the 16
	detectors acquiring data during each scan. */
	if (dim_sizes[0] % 16) return error(ERR_IMAGE_LINES);

	/* sixteen lines per full VIIRS scan */
	hinfo->nscans = dim_sizes[0] / 16;


	/* M13 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M13",
		&hinfo->sds_index_QF13, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M13 radiance */
	if ( (st = init_input_sds(sd_id_vmae, "Radiance_M13",
		&hinfo->sds_index_L13, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M15 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M15",
		&hinfo->sds_index_QF15, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M15 */
	if ( (st = init_input_sds(sd_id_vmae, "BrightnessTemperature_M15",
		&hinfo->sds_index_M15, DFNT_UINT16, dim_sizes,
		&hinfo->ScaleT15, &hinfo->OffsetT15, 0)) ) return error(st);

	/* M16 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M16",
		&hinfo->sds_index_QF16, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M16 */
	if ( (st = init_input_sds(sd_id_vmae, "BrightnessTemperature_M16",
		&hinfo->sds_index_M16, DFNT_UINT16, dim_sizes,
		&hinfo->ScaleT16, &hinfo->OffsetT16, 0)) ) return error(st);

	/* M5 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M5",
		&hinfo->sds_index_QF5, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M5 (red) */
	if ( (st = init_input_sds(sd_id_vmae, "Reflectance_M5",
		&hinfo->sds_index_M5, DFNT_UINT16, dim_sizes,
		&hinfo->ScaleR5, &hinfo->OffsetR5,
		OPTIONAL_SCALE_OFFSET)) ) return error(st);

	/* M7 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M7",
		&hinfo->sds_index_QF7, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M7 (NIR) */
	if ( (st = init_input_sds(sd_id_vmae, "Reflectance_M7",
		&hinfo->sds_index_M7, DFNT_UINT16, dim_sizes,
		&hinfo->ScaleR7, &hinfo->OffsetR7,
		OPTIONAL_SCALE_OFFSET)) ) return error(st);

	/* M11 QA */
	if ( (st = init_input_sds(sd_id_vmae, "QF1_VIIRSMBANDSDR_M11",
		&hinfo->sds_index_QF11, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);

	/* M11 (SWIR) */
	if ( (st = init_input_sds(sd_id_vmae, "Reflectance_M11",
		&hinfo->sds_index_M11, DFNT_UINT16, dim_sizes,
		&hinfo->ScaleR11, &hinfo->OffsetR11,
		OPTIONAL_SCALE_OFFSET)) ) return error(st);



	/* geolocation */
	if ( (st = init_input_sds(sd_id_vmae, "Latitude",
		&hinfo->sds_index_lat, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);
	if ( (st = init_input_sds(sd_id_vmae, "Longitude",
		&hinfo->sds_index_lon, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);
	if ( (st = init_input_sds(sd_id_vmae, "SolarZenithAngle",
		&hinfo->sds_index_solzen, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);
	if ( (st = init_input_sds(sd_id_vmae, "SolarAzimuthAngle",
		&hinfo->sds_index_solaz, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);
	if ( (st = init_input_sds(sd_id_vmae, "SatelliteZenithAngle",
		&hinfo->sds_index_viewzen, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);
	if ( (st = init_input_sds(sd_id_vmae, "SatelliteAzimuthAngle",
		&hinfo->sds_index_viewaz, DFNT_FLOAT32, dim_sizes,
		NULL, NULL, 0)) ) return error(st);


	/* ---------- NPP_QSLWMIP_L2 ---------- */

	if ( (st = init_input_sds(sd_id_lc, "QSTLWM",
		&hinfo->sds_index_lc, DFNT_UINT8, dim_sizes,
		NULL, NULL, 0)) ) return error(st);


	/* Use VMAE to identify satellite.  NOTE: Currently this function
	is merely a stub that ignores the SD ID parameter and returns
	a fixed result of SAT_NPP in hinfo->sat. */
	if ( (st = identify_satellite(sd_id_vmae, &hinfo->sat)) ) return st;


	/* Use VMAE to obtain the gring */
	if ( (st = read_gring(sd_id_vmae, hinfo->gringLatLonStr)) )
		return st;


	return 0;
}



static int init_input_sds(int32 sd_id, char *sds_name, int32 *sds_index_ptr,
	int32 expected_data_type, int32 dim_sizes[MAX_VAR_DIMS],
	float32 *scale, float32 *offset, int optional_scale)
{
	int32 sds_index, sds_id, rank, data_type, nattr;
	char dummy_name[MAX_NC_NAME];
	int st;

	if ((sds_index = SDnametoindex(sd_id, sds_name)) == FAIL)
		return ERR_HDF;
	if ((sds_id = SDselect(sd_id, sds_index)) == FAIL) return ERR_HDF;

	if (SDgetinfo(sds_id, dummy_name, &rank, dim_sizes, &data_type,
		&nattr) == FAIL) return ERR_HDF;

	if (verbose)
		printf("name = %24s, rank = %2d, data type = %2d  dims: %s\n",
			sds_name, rank, data_type, DimStr(rank, dim_sizes));

	/* check rank and data type */
	if (rank != 2) return ERR_DATA_RANK;
	if (data_type != expected_data_type) return ERR_DATA_TYPE;

	if (scale) {
		if ( (st = read_scale_offset(sds_id, scale, offset,
			optional_scale)) ) return st;

		if (verbose) printf("scale: %.6f  offset: %.6f\n",
			*scale, *offset);
		}

	if (SDendaccess(sds_id) == FAIL) return ERR_HDF;

	*sds_index_ptr = sds_index;
	return 0;
}




static char *DimStr(int32 rank, int32 *dim_sizes)
{
	int i;
	char numstr[8];
	static char s[128];

	s[0] = '\0';
	for (i = 0; i < rank; i++) {
		(void) sprintf(numstr, "%6d", dim_sizes[i]);
		(void) strcat(s, numstr);
		}

	return s;
}




/*
Read a single VIIRS scan from M-band input file.

	hinfo		HDF input file info.
	iscan		scan number (0 <= iscan < hinfo.nscans)
	scan		Scan structure in which data should be stored

Returns 0 if no errors, non-zero error code otherwise.
*/
int ReadScan(HDF_InputInfo *hinfo, int iscan, Scan *scan)
{
	int32 start[2];
	int32 edges[2];

	uint16 buff[LINES_PER_SCAN_750M][SAMPLES_750M];


	start[0] = LINES_PER_SCAN_750M * iscan;
	edges[0] = LINES_PER_SCAN_750M;	/* read 16 scan lines (i.e. one complete scan) */

	start[1] = 0;
	edges[1] = SAMPLES_750M;


	/* ----- VMAE ----- */

	/* M13 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF13,
		start, edges, (void *) scan->QF13)) return error(ERR_HDF);

	/* M13 brightness temperature */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M13,
		start, edges, (void *) scan->T13)) return error(ERR_HDF);

	/* M13 radiance */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_L13,
		start, edges, (void *) scan->L13)) return error(ERR_HDF);

	/* M15 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF15,
		start, edges, (void *) scan->QF15)) return error(ERR_HDF);

	/* M15 scaled brightness temperature */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M15,
		start, edges, (void *) buff)) return error(ERR_HDF);
	unscale(buff, scan->T15, hinfo->ScaleT15, hinfo->OffsetT15);

	/* M16 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF16,
		start, edges, (void *) scan->QF16)) return error(ERR_HDF);

	/* M16 scaled brightness temperature */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M16,
		start, edges, (void *) buff)) return error(ERR_HDF);
	unscale(buff, scan->T16, hinfo->ScaleT16, hinfo->OffsetT16);

	/* M5 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF5,
		start, edges, (void *) scan->QF5)) return error(ERR_HDF);

	/* M5 scaled reflectance */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M5,
		start, edges, (void *) buff)) return error(ERR_HDF);
	unscale(buff, scan->R5, hinfo->ScaleR5, hinfo->OffsetR5);

	/* M7 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF7,
		start, edges, (void *) scan->QF7)) return error(ERR_HDF);

	/* M7 scaled reflectance */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M7,
		start, edges, (void *) buff)) return error(ERR_HDF);
	unscale(buff, scan->R7, hinfo->ScaleR7, hinfo->OffsetR7);

	/* M11 QA */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_QF11,
		start, edges, (void *) scan->QF11)) return error(ERR_HDF);

	/* M11 scaled reflectance */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_M11,
		start, edges, (void *) buff)) return error(ERR_HDF);
	unscale(buff, scan->R11, hinfo->ScaleR11, hinfo->OffsetR11);

	/* latitude */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_lat,
		start, edges, (void *) scan->lat)) return error(ERR_HDF);

	/* longitude */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_lon,
		start, edges, (void *) scan->lon)) return error(ERR_HDF);

	/* view zenith angle */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_viewzen,
		start, edges, (void *) scan->ViewZenAng)) return error(ERR_HDF);

	/* solar zenith angle */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_solzen,
		start, edges, (void *) scan->SolZenAng)) return error(ERR_HDF);

	/* view azimuth angle */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_viewaz,
		start, edges, (void *) scan->ViewAzAng)) return error(ERR_HDF);

	/* solar azimuth angle */
	if (read_band_scan(hinfo->sd_id_vmae, hinfo->sds_index_solaz,
		start, edges, (void *) scan->SolAzAng)) return error(ERR_HDF);

	/* ----- QSLWMIP surface type ----- */
	if (read_band_scan(hinfo->sd_id_lc, hinfo->sds_index_lc,
		start, edges, (void *) scan->landsea)) return error(ERR_HDF);

	return 0;
}



static int read_band_scan(int32 sd_id, int32 sds_index, int32 start[2],
	int32 edges[2], void *buff)
{
	int32 sds_id;

	if ((sds_id = SDselect(sd_id, sds_index)) == FAIL) return ERR_HDF;
	if (SDreaddata(sds_id, start, NULL, edges, buff) == FAIL)
		return ERR_HDF;
	if (SDendaccess(sds_id) == FAIL) return ERR_HDF;

	return 0;
}



/* Unscale scaled brightness temperatures and reflectances into physical values.
   Missing data will be flagged with a value of -1. */
static void unscale(uint16 ScaledVal[LINES_PER_SCAN_750M][SAMPLES_750M],
	float32 UnscaledVal[LINES_PER_SCAN_750M][SAMPLES_750M],
	float scale, float offset)
{
	int i, j;
	const float missval = -1.0F;

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (ScaledVal[i][j] <= UINT16_FILL_TEST)
				UnscaledVal[i][j] =
					scale * ScaledVal[i][j] + offset;
			else
				UnscaledVal[i][j] = missval;
			}
		}
}




/*
This is a stub function used to identify the particular platform on which
the VIIRS sensor resides.  For the time being the only possible result
outcome can be SAT_NPP, corresponding to the NPP satellite.
The function will eventually have to be modified to distinguish
between NPP, JPSS-1, JPSS-2, etc.

Return value is 0 for success, non-zero if error.  Currently only a
successful outcome is possible.  The non-zero error result is a leftover
from the original version of this function, which distinguished between
the Terra and Aqua satellites for MODIS.  Because metadata from the
MODIS geolocation product was used to make the determination, we had to
allow for the possibility of an I/O error occurring in the process.
(In that case "sat" was set to SAT_UNKNOWN, and a non-zero error code
was returned.)
*/
static int identify_satellite(int32 sd_id_vmae, enum Satellite *sat)
{
	*sat = SAT_NPP;

	return 0;	/* success */
}



/*
For some strange reason the Scale and Offset attributes are not present in
the reflectance SDSs in nighttime granules.  This function
will handle such cases by silently returning a scale of 1.0 and an offset of
0.0 when these attributes are missing AND the flag "optional" is set, otherwise
the function will return a non-zero error code when the attributes are missing.
*/
static int read_scale_offset(int32 sds_id, float32 *scale,
	float32 *offset, int optional)
{
	int32 attr_index;
	int32 n_values;
	int32 data_type;
	char attr_name[MAX_NC_NAME];


	if ((attr_index = SDfindattr(sds_id, "Scale")) == FAIL) {
		if (optional) *scale = 1.0F; else return ERR_MISSING_ATTR;
		}
	else {
		if (SDattrinfo(sds_id, attr_index, attr_name, &data_type,
			&n_values)) return ERR_HDF;
		if (data_type != DFNT_FLOAT32) return ERR_DATA_TYPE;
		if (SDreadattr(sds_id, attr_index, scale)) return ERR_HDF;
		}


	if ((attr_index = SDfindattr(sds_id, "Offset")) == FAIL) {
		if (optional) *offset = 0.0F; else return ERR_MISSING_ATTR;
		}
	else {
		if (SDattrinfo(sds_id, attr_index, attr_name, &data_type,
			&n_values)) return ERR_HDF;
		if (data_type != DFNT_FLOAT32) return ERR_DATA_TYPE;
		if (SDreadattr(sds_id, attr_index, (void *) offset))
			return ERR_HDF;
		}

	return 0;
}





int read_gring(int32 sds_id, char *gringLatLonStr[2])
{
	int32 attr_index;
	const char *attr_name;
	const char* attr_name_glat;
	const char* attr_name_glon;
	char attr_name_buf[MAX_NC_NAME];
	int32 data_type;
	int32 n_values;
	int i;


	attr_name_glat=GRINGLATATTNAM1;
	attr_name_glon=GRINGLONATTNAM1;

	attr_name=attr_name_glat;
	for(i = 0; i < 2; i++){
		if ((attr_index = SDfindattr(sds_id, attr_name)) == FAIL)
			return error(ERR_HDF);
		if (SDattrinfo(sds_id, attr_index, attr_name_buf, &data_type,
			&n_values)) return error(ERR_HDF);
		if (data_type != DFNT_CHAR) return error(ERR_DATA_TYPE);

		/* allocate extra character for terminating null */
		(gringLatLonStr)[i] = (char *) calloc(n_values + 1,
			sizeof(char));
		if ( !((gringLatLonStr)[i]) ) return error(ERR_F_MEM);

		if (SDreadattr(sds_id, attr_index,
			(void *) ((gringLatLonStr)[i]))) return error(ERR_HDF);

		/* set up for second iteration */
		attr_name=attr_name_glon;
		}

	return ERR_NONE;
}

