/*
!C******************************************************************
!Description:
	Functions to handle output to HDF files.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio, July 2011

!Design Notes:

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "mfhdf.h"

#include "const.h"
#include "types.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "output.h"
#include "error.h"
#include "misc.h"
#include "version.h"



extern int write_tb, write_ref;

static int create_thermal_band(int32 sd_id, char *sds_name, char *viirs_name,
	int32 dim_sizes[2], char *dimname1, char *dimname2, HDF_CHUNK_DEF cdef);
static int create_ref_band(int32 sd_id, int channel, int32 dim_sizes[2],
	char *dimname1, char *dimname2, HDF_CHUNK_DEF cdef);
#ifdef WRITE_DIAGNOSTIC
static int create_diagnostic_layers(int32 sd_id, int nscans, int *sds_index,
	HDF_OutputInfo *hinfo);
#endif

static int set_dimensions(int32 sds_id, char *dimname1, char *dimname2);

static void scale_ref(float R[LINES_PER_SCAN_750M][SAMPLES_750M],
	uint16 ubuff[LINES_PER_SCAN_750M][SAMPLES_750M]);

static void scale_tb(float32 Tb[LINES_PER_SCAN_750M][SAMPLES_750M],
	uint16 ubuff[LINES_PER_SCAN_750M][SAMPLES_750M]);

static int write_band_scan(int32 sd_id, int32 sds_index, int32 start[2],
	int32 edges[2], void *buff);

static char *satstr(enum Satellite sat);

static int init_geo_common_output(int32 sds_id, char *dimname1,
	char *dimname2);


static const char *GRINGLATATTNAM1="Unagg_GRingLatitude";
static const char *GRINGLONATTNAM1="Unagg_GRingLongitude";




int InitOutputHDF(int32 sd_id_fire, int nscans, int write_geo, HDF_OutputInfo *hinfo)
{
	int32 sds_id;

	int32 dim_sizes[2];
	char *dimname1, *dimname2;

	int sds_index = 0;

	int32 n_values;
	uint8 ubuff[2];

	char *units;
	char *legend;

	HDF_CHUNK_DEF cdef;


	/* ========== Level 2 Fire Product ========== */
	hinfo->sd_id_fire = sd_id_fire;

	dim_sizes[0] = nscans * LINES_PER_SCAN_750M;
	dim_sizes[1] = SAMPLES_750M;

	dimname1 = "number_of_scan_lines";
	dimname2 = "pixels_per_scan_line";

	cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = LINES_PER_SCAN_750M;
	cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = SAMPLES_750M;
	cdef.comp.comp_type = COMP_CODE_DEFLATE;
	cdef.comp.cinfo.deflate.level = 4;


	/* ----- fire mask ----- */

	hinfo->sds_index_mask = sds_index++;
	if ((sds_id = SDcreate(sd_id_fire, "fire mask", DFNT_UINT8, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);

	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if (set_dimensions(sds_id, dimname1, dimname2))
		return error(ERR_HDF);

	ubuff[0] = 0;
	ubuff[1] = 9;
	n_values = 2;
	if (SDsetattr(sds_id, "valid_range", DFNT_UINT8, n_values, ubuff))
		return error(ERR_HDF);

legend = "Classes:\n"
	"0 missing input data\n"
	"1 not processed (obsolete)\n"
	"2 not processed (obsolete)\n"
	"3 non-fire water\n4 cloud\n5 non-fire land\n"
	"6 unknown\n7 fire (low confidence)\n"
	"8 fire (nominal confidence)\n9 fire (high confidence)";

	n_values = strlen(legend);
	if (SDsetattr(sds_id, "legend", DFNT_CHAR8, n_values, legend))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	/* ----- QA ----- */

	hinfo->sds_index_qa = sds_index++;
	if ((sds_id = SDcreate(sd_id_fire, "algorithm QA", DFNT_UINT32, 2,
		dim_sizes)) == FAIL)
		return error(ERR_HDF);

	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if (set_dimensions(sds_id, dimname1, dimname2))
		return error(ERR_HDF);

	units = "bit field";
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


#ifdef UNUSED
	/* extended fire mask (C6) */
	hinfo->sds_index_emask = sds_index++;
	if ((sds_id = SDcreate(sd_id_fire, "extended fire mask", DFNT_UINT8, 2, dim_sizes)) == FAIL)
		return error(ERR_HDF);

	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if (set_dimensions(sds_id, dimname1, dimname2))
		return error(ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
#endif

	if (write_tb) {
		hinfo->sds_index_T13 = sds_index++;
		if (create_thermal_band(sd_id_fire, "T13", "M13", dim_sizes,
			dimname1, dimname2, cdef)) return (ERR_HDF);

		hinfo->sds_index_T15 = sds_index++;
		if (create_thermal_band(sd_id_fire, "T15", "M15", dim_sizes,
			dimname1, dimname2, cdef)) return (ERR_HDF);
		}
	else
		hinfo->sds_index_T13 = hinfo->sds_index_T15 =
			hinfo->sds_index_T16 = -1;

	if (write_ref) {
		hinfo->sds_index_R7 = sds_index++;
		if (create_ref_band(sd_id_fire, 7, dim_sizes,
			dimname1, dimname2, cdef)) return (ERR_HDF);

		hinfo->sds_index_R11 = sds_index++;
		if (create_ref_band(sd_id_fire, 11, dim_sizes,
			dimname1, dimname2, cdef)) return (ERR_HDF);
		}
	else
		hinfo->sds_index_R7 = hinfo->sds_index_R11 = -1;

	if (write_geo) {
		hinfo->sds_index_lat = sds_index++;
		if ((sds_id = SDcreate(sd_id_fire, "latitude", DFNT_FLOAT32, 2,
			dim_sizes)) == FAIL) return (ERR_HDF);
		if (init_geo_common_output(sds_id, dimname1, dimname2))
			return (ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);

		hinfo->sds_index_lon = sds_index++;
		if ((sds_id = SDcreate(sd_id_fire, "longitude", DFNT_FLOAT32, 2,
			dim_sizes)) == FAIL) return (ERR_HDF);
		if (init_geo_common_output(sds_id, dimname1, dimname2))
			return (ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);
		}
	else
		hinfo->sds_index_lat = hinfo->sds_index_lon = -1;

#ifdef WRITE_DIAGNOSTIC
	if (create_diagnostic_layers(sd_id_fire, nscans, &sds_index, hinfo)) {
		return (ERR_HDF);
		}
#endif

	return 0;
}



static int create_thermal_band(int32 sd_id, char *sds_name, char *viirs_name,
	int32 dim_sizes[2], char *dimname1, char *dimname2, HDF_CHUNK_DEF cdef)
{
	int32 sds_id;
	int32 len;
	const char *units = "kelvins";
	/* increase size of long_name array if strlen(viirs_name) > 40 */
	char long_name[64];
	int16 missing_value = 0;
	float32 scale_factor = 0.1F;


	if ((sds_id = SDcreate(sd_id, sds_name, DFNT_UINT16, 2,
		dim_sizes)) == FAIL) return (ERR_HDF);

	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);

	(void) sprintf(long_name, "%s brightness temperature", viirs_name);
	len = strlen(long_name);
	if (SDsetattr(sds_id, "long_name", DFNT_CHAR8, len, long_name)) return (ERR_HDF);

	len = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, len, units)) return (ERR_HDF);
	if (SDsetattr(sds_id, "missing_value", DFNT_INT16, 1, &missing_value))
		return (ERR_HDF);
	if (SDsetattr(sds_id, "scale_factor", DFNT_FLOAT32, 1, &scale_factor))
		return (ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);

	return 0;
}




static int create_ref_band(int32 sd_id, int channel, int32 dim_sizes[2],
	char *dimname1, char *dimname2, HDF_CHUNK_DEF cdef)
{
	int32 sds_id;
	int32 len;
	const char *units = "dimensionless";
	char sds_name[8];	/* "R5", "R11", etc. */
	char long_name[24];	/* "M5 reflectance", "M11 reflectance", etc. */
	int16 missing_value = 0;
	float32 scale_factor = 0.001F;


	(void) sprintf(sds_name, "R%d", channel);
	(void) sprintf(long_name, "M%d reflectance", channel);


	if ((sds_id = SDcreate(sd_id, sds_name, DFNT_UINT16, 2,
		dim_sizes)) == FAIL) return (ERR_HDF);

	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);

	len = strlen(long_name);
	if (SDsetattr(sds_id, "long_name", DFNT_CHAR8, len, long_name)) return (ERR_HDF);
	len = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, len, units)) return (ERR_HDF);
	if (SDsetattr(sds_id, "missing_value", DFNT_INT16, 1, &missing_value))
		return (ERR_HDF);
	if (SDsetattr(sds_id, "scale_factor", DFNT_FLOAT32, 1, &scale_factor))
		return (ERR_HDF);

	if (SDendaccess(sds_id) == FAIL) return (ERR_HDF);

	return 0;
}



static int init_geo_common_output(int32 sds_id, char *dimname1, char *dimname2)
{
	int32 dim_id;
	int32 ulen;
	char *units;

	HDF_CHUNK_DEF cdef;


	units = "degrees";
	ulen = strlen(units);


	cdef.chunk_lengths[0] = cdef.comp.chunk_lengths[0] = LINES_PER_SCAN_750M;
	cdef.chunk_lengths[1] = cdef.comp.chunk_lengths[1] = SAMPLES_750M;
	cdef.comp.comp_type = COMP_CODE_DEFLATE;
	cdef.comp.cinfo.deflate.level = 2;
	if (SDsetchunk(sds_id, cdef, HDF_CHUNK | HDF_COMP) == FAIL)
		return error(ERR_HDF);

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL) return (ERR_HDF);
	if (SDsetdimname(dim_id, dimname1) == FAIL) return (ERR_HDF);
	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return (ERR_HDF);
	if (SDsetdimname(dim_id, dimname2) == FAIL) return (ERR_HDF);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, ulen, units)) return (ERR_HDF);

	return 0;
}


#ifdef WRITE_DIAGNOSTIC
/* write diagnostic layers containing zone statistics */
static int create_diagnostic_layers(int32 sd_id, int nscans, int *sds_index,
	HDF_OutputInfo *hinfo)
{
	int32 sds_id;

	int32 dim_sizes[2];
	char *dimname1 = "number_of_scans";
	char *dimname2 = "pixels_per_scan_line";


	int32 n_values;

	const char *units = "kelvins";


printf("sds_index = %d\n", *sds_index);


	dim_sizes[0] = nscans;
	dim_sizes[1] = SAMPLES_750M;


	hinfo->sds_index_LandMedianT13 = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "MedianT13_land", DFNT_FLOAT32, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_LandMedianDT = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "MedianDT_land", DFNT_FLOAT32, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_LandZonePix = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "ZonePix_land", DFNT_UINT16, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_WaterMedianT13 = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "MedianT13_water", DFNT_FLOAT32, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_WaterMedianDT = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "MedianDT_water", DFNT_FLOAT32, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	n_values = strlen(units);
	if (SDsetattr(sds_id, "units", DFNT_CHAR8, n_values, units))
		return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);


	hinfo->sds_index_WaterZonePix = (*sds_index)++;
	if ((sds_id = SDcreate(sd_id, "ZonePix_water", DFNT_UINT16, 2,
		dim_sizes)) == FAIL) return error(ERR_HDF);
	if (set_dimensions(sds_id, dimname1, dimname2)) return error(ERR_HDF);
	if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

	return 0;
}
#endif




static int set_dimensions(int32 sds_id, char *dimname1, char *dimname2)
{
	int32 dim_id;

	if ((dim_id = SDgetdimid(sds_id, 0)) == FAIL) return ERR_HDF;
	if (SDsetdimname(dim_id, dimname1) == FAIL) return ERR_HDF;
	if ((dim_id = SDgetdimid(sds_id, 1)) == FAIL) return ERR_HDF;
	if (SDsetdimname(dim_id, dimname2) == FAIL) return ERR_HDF;

	return 0;
}



/*
Scan pointer is passed in case one of the buffered quantities needs to be
added to L2 fire product.  It is currently only used 
to write brightness temperature arrays.
*/
int WriteScan(HDF_OutputInfo *hinfo, int iscan, OutScan *outscan, Scan *scan,
	ScanThreshPF *pfthresh)
{
	int32 sds_id;
	int32 start[2];
	int32 edges[2];

	uint16 ubuff[LINES_PER_SCAN_750M][SAMPLES_750M];


	start[0] = LINES_PER_SCAN_750M * iscan;
	edges[0] = LINES_PER_SCAN_750M;

	start[1] = 0;
	edges[1] = SAMPLES_750M;

	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_mask, start,
		edges, (void *) outscan->FireMask)) return error(ERR_HDF);

	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_qa, start,
		edges, (void *) outscan->FireQA)) return error(ERR_HDF);


#ifdef UNUSED
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_emask, start,
		edges, (void *) outscan->ExtFireMask)) return error(ERR_HDF);
#endif


	if (write_tb) {
		/* M13 */
		scale_tb(scan->T13, ubuff);
		if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_T13,
			start, edges, (void *) ubuff)) return error(ERR_HDF);

		/* M15 */
		scale_tb(scan->T15, ubuff);
		if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_T15,
			start, edges, (void *) ubuff)) return error(ERR_HDF);
		}

	if (write_ref) {
		/* M7 */
		scale_ref(scan->R7, ubuff);
		if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_R7,
			start, edges, (void *) ubuff)) return error(ERR_HDF);

		/* M11 */
		scale_ref(scan->R11, ubuff);
		if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_R11,
			start, edges, (void *) ubuff)) return error(ERR_HDF);
		}


	/* write geolocation if requested */
	if (hinfo->sds_index_lat >= 0) {
		if ((sds_id = SDselect(hinfo->sd_id_fire,
			hinfo->sds_index_lat)) == FAIL) return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) scan->lat)
			== FAIL) return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);

		if ((sds_id = SDselect(hinfo->sd_id_fire,
			hinfo->sds_index_lon)) == FAIL) return error(ERR_HDF);
		if (SDwritedata(sds_id, start, NULL, edges, (void *) scan->lon)
			== FAIL) return error(ERR_HDF);
		if (SDendaccess(sds_id) == FAIL) return error(ERR_HDF);
		}

#ifdef WRITE_DIAGNOSTIC
	/* write scan statistics */
	start[0] = iscan;
	edges[0] = 1;

	start[1] = 0;
	edges[1] = SAMPLES_750M;

	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_LandMedianT13,
		start, edges, (void *) scan->LandMedianT13))
		return error(ERR_HDF);
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_LandMedianDT,
		start, edges, (void *) scan->LandMedianDT))
		return error(ERR_HDF);
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_LandZonePix,
		start, edges, (void *) scan->LandZonePix))
		return error(ERR_HDF);
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_WaterMedianT13,
		start, edges, (void *) scan->WaterMedianT13))
		return error(ERR_HDF);
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_WaterMedianDT,
		start, edges, (void *) scan->WaterMedianDT))
		return error(ERR_HDF);
	if (write_band_scan(hinfo->sd_id_fire, hinfo->sds_index_WaterZonePix,
		start, edges, (void *) scan->WaterZonePix))
		return error(ERR_HDF);
#endif


	return 0;
}



static int write_band_scan(int32 sd_id, int32 sds_index, int32 start[2],
	int32 edges[2], void *buff)
{
	int32 sds_id;

	if ((sds_id = SDselect(sd_id, sds_index)) == FAIL)
		return ERR_HDF;
	if (SDwritedata(sds_id, start, NULL, edges, buff) == FAIL)
		return ERR_HDF;
	if (SDendaccess(sds_id) == FAIL) return ERR_HDF;

	return 0;
}



static void scale_tb(float32 Tb[LINES_PER_SCAN_750M][SAMPLES_750M],
	uint16 ubuff[LINES_PER_SCAN_750M][SAMPLES_750M])
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (Tb[i][j] > 0.0F)
				ubuff[i][j] = (uint16) (10.0F * Tb[i][j]);
			else
				ubuff[i][j] = 0;
			}
		}
}




static void scale_ref(float R[LINES_PER_SCAN_750M][SAMPLES_750M],
	uint16 ubuff[LINES_PER_SCAN_750M][SAMPLES_750M])
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			if (R[i][j] > 0.0F)
				ubuff[i][j] = (uint16) (1000.0F * R[i][j] + 0.5F);
			else
				ubuff[i][j] = 0;
			}
		}
}



int write_fire_attributes(int32 sd_id, FireCounts *counts)
{
	const int32 n_values = 1;

	if (SDsetattr(sd_id, "FirePix", DFNT_INT32, n_values,
		&counts->NumFirePix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "LandFirePix", DFNT_INT32, n_values,
		&counts->NumLandFirePix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "WaterFirePix", DFNT_INT32, n_values,
		&counts->NumWaterFirePix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "MissingPix", DFNT_INT32, n_values,
		&counts->NumMissingPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "LandPix", DFNT_INT32, n_values,
		&counts->NumLandPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "WaterPix", DFNT_INT32, n_values,
		&counts->NumWaterPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "WaterAdjacentFirePix", DFNT_INT32, n_values,
		&counts->NumWaterAdjFirePix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "CloudAdjacentFirePix", DFNT_INT32, n_values,
		&counts->NumCloudAdjFirePix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "UnknownLandPix", DFNT_INT32, n_values,
		&counts->NumUnknownLandPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "UnknownWaterPix", DFNT_INT32, n_values,
		&counts->NumUnknownWaterPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "LandCloudPix", DFNT_INT32, n_values,
		&counts->NumLandCloudPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "WaterCloudPix", DFNT_INT32, n_values,
		&counts->NumWaterCloudPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "GlintPix", DFNT_INT32, n_values,
		&counts->NumGlintPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "GlintRejectedPix", DFNT_INT32, n_values,
		&counts->NumGlintRejectPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "CoastRejectedLandPix", DFNT_INT32, n_values,
		&counts->NumCoastRejectLandPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "HotSurfRejectedPix", DFNT_INT32, n_values,
		&counts->NumHotRejectPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "ClearingRejectedPix", DFNT_INT32, n_values,
		&counts->NumClearingRejectPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "CoastRejectedWaterPix", DFNT_INT32, n_values,
		&counts->NumCoastRejectWaterPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "DayPix", DFNT_INT32, n_values,
		&counts->NumDayPix)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "NightPix", DFNT_INT32, n_values,
		&counts->NumNightPix)) return error(ERR_HDF);

	if (SDsetattr(sd_id, "MissingRadPix", DFNT_INT32, n_values,
		&counts->NumMissingRad)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "MissingGeoPix", DFNT_INT32, n_values,
		&counts->NumMissingGeo)) return error(ERR_HDF);
	if (SDsetattr(sd_id, "TrimmedPix", DFNT_INT32, n_values,
		&counts->NumTrimPix)) return error(ERR_HDF);


	return 0;
}




int write_generic_attributes(int32 sd_id, char *infile_vmae, char *infile_lc,
	enum Satellite sat, char *gringLatLonStr[2])
{
	int32 n_values;
	char *s;

	s = satstr(sat);
	n_values = strlen(s);
	if (SDsetattr(sd_id, "Satellite", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	s = PROCESS_VERSION_NUMBER;
	n_values = strlen(s);
	if (SDsetattr(sd_id, "ProcessVersionNumber", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	n_values = strlen(infile_vmae);
	if (SDsetattr(sd_id, "VMAE input file", DFNT_CHAR8, n_values,
		infile_vmae)) return error(ERR_HDF);

	n_values = strlen(infile_lc);
	if (SDsetattr(sd_id, "QSLWMIP input file", DFNT_CHAR8, n_values,
		infile_lc)) return error(ERR_HDF);

	s = make_uname();
	n_values = strlen(s);
	if (SDsetattr(sd_id, "SystemID", DFNT_CHAR8, n_values, s))
		return error(ERR_HDF);

	n_values = strlen(gringLatLonStr[0]);
	if (SDsetattr(sd_id, GRINGLATATTNAM1, DFNT_CHAR8, n_values,
		gringLatLonStr[0])) return error(ERR_HDF);

	n_values = strlen(gringLatLonStr[1]);
	if (SDsetattr(sd_id, GRINGLONATTNAM1, DFNT_CHAR8, n_values,
		gringLatLonStr[1])) return error(ERR_HDF);

	return 0;
}



/* return pointer to static string naming satellite */
static char *satstr(enum Satellite sat)
{
	if (sat == SAT_NPP) return "NPP";
	return "?";
}

