/*
!C******************************************************************
!Description:

	Functions related to processing of each granule.

!Revision History:
!Team-unique Header:
!References and Credits:

	L. Giglio, September 2011
	W. Walsh, May 2014

!END*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "fire.h"
#include "detfire.h"
#include "firetab.h"
#include "output.h"
#include "input.h"
#include "process.h"
#include "cmg.h"
#include "zstats.h"
#include "error.h"
#ifdef DIAGNOSE_SDR
#include "monitor.h" 
#endif

static void clear_scan(Scan *scan);
static void update_cmg(OutScan *outscan, Scan *scan);



/* ---------- global flags ---------- */

extern int verbose;
extern int write_coarse;
extern int write_cmg;



int ProcGranule(HDF_InputInfo *hinfo, HDF_OutputInfo *hfire,
	FireCounts *counts, int skip_water)
{
	int i;
	int iscan;
	int st;
	Scan *PrevPtr, *CurrPtr, *NextPtr;
	Scan *save;

	Scan *scanptr[3];	/*  0 = previous, 1 = current, 2 = next */

	OutScan *outscan;

	int first_line;

	ScanThreshPF *pfthresh;



	/* allocate structure to hold potential fire thresholds for scan */
	if ( !(pfthresh = (ScanThreshPF *) malloc(sizeof(ScanThreshPF))) )
		return error(ERR_F_MEM);

	/* allocate buffer for output */
	if ( !(outscan = (OutScan *) malloc(sizeof(OutScan))) )
		return error(ERR_F_MEM);

	/* allocate triple buffer */
	for (i = 0; i < 3; i++) {
		if ( !(scanptr[i] = (Scan *) malloc(sizeof(Scan))) )
			return error(ERR_F_MEM);
		}

	PrevPtr = scanptr[0];
	CurrPtr = scanptr[1];
	NextPtr = scanptr[2];

	if ( (st = init_zstats()) ) return st;

	/* zero pixel counts */
	(void) memset(counts, 0, sizeof(FireCounts));


	/* handle zero scans */
	if (hinfo->nscans < 1) {
		if ( (st = write_fire_attributes(hfire->sd_id_fire, counts)) )
			return st;

#ifdef NOT_IMP
		if (write_coarse)
			if ( (st = write_fire_attributes(hfire->sd_id_coarse,
				counts)) ) return st;
#endif

		free(pfthresh);
		free(outscan);

		/* free triple buffer */
		for (i = 0; i < 3; i++) free(scanptr[i]);

		free_zstats();

		return 0;
		}

	new_fire_pixel_table();


	/* clear previous buffer */
	clear_scan(PrevPtr);

	/* fill current buffer */
	st = ReadScan(hinfo, 0, CurrPtr);
	if (st) return st;

	/* screen current buffer */
	prescreen(CurrPtr, skip_water);
#ifdef DIAGNOSE_SDR
	update_monitor(CurrPtr);
#endif

	first_line = 0;
	for (iscan = 1; iscan < hinfo->nscans; iscan++) {
		if (verbose) if (!(iscan % 10)) printf("%d\n", iscan);

		/* fill next buffer */
		st = ReadScan(hinfo, iscan, NextPtr);
		if (st) return st;

		/* screen next buffer */
		prescreen(NextPtr, skip_water);
#ifdef DIAGNOSE_SDR
		update_monitor(NextPtr);
#endif

		/* ----- detect fires in current scan (CurrPtr) ----- */
		compute_zone_stats(PrevPtr, CurrPtr, NextPtr);

		/* pre-compute potential fire thresholds for each sample */
		compute_scan_PF_thresholds(CurrPtr, pfthresh);

		/* flag potential fire pixels in current scan (the contextual
		detection will only be performed on this buffer) */
		flag_potential_fires(CurrPtr, pfthresh);

		/* unlike potential fire pixels, we must flag
		background fire pixels in all three buffers because the
		contextual detection phase will use observations from all of
		them */
		flag_background_fires(PrevPtr);
		flag_background_fires(CurrPtr);
		flag_background_fires(NextPtr);

		st = detect(PrevPtr, CurrPtr, NextPtr, first_line, skip_water,
			pfthresh, outscan, counts);
		if (st) return st;

		if ( (st = WriteScan(hfire, iscan - 1, outscan, CurrPtr,
			pfthresh)) ) return st;

		if (write_cmg) update_cmg(outscan, CurrPtr);

		save = PrevPtr;
		PrevPtr = CurrPtr;
		CurrPtr = NextPtr;
		NextPtr = save;


		/*
		At this point scan buffer pointed to by NextPtr contains data
		from an earlier scan that will be overwritten.
		*/

		first_line += LINES_PER_SCAN_750M;
		}

	clear_scan(NextPtr);


	/* do detection on current buffer */
	compute_zone_stats(PrevPtr, CurrPtr, NextPtr);

	compute_scan_PF_thresholds(CurrPtr, pfthresh);

	flag_potential_fires(CurrPtr, pfthresh);

	flag_background_fires(PrevPtr);
	flag_background_fires(CurrPtr);

	st = detect(PrevPtr, CurrPtr, NextPtr, first_line, skip_water,
		pfthresh, outscan, counts);
	if (st) return st;

	if ( (st = WriteScan(hfire, iscan - 1, outscan, CurrPtr,
		pfthresh)) ) return st;
	if (write_cmg) update_cmg(outscan, CurrPtr);

	free(pfthresh);
	free(outscan);

	/* free triple buffer */
	for (i = 0; i < 3; i++) free(scanptr[i]);

	free_zstats();

	st = write_fire_pixel_table(hfire->sd_id_fire, write_cmg);
	if (st) return st;

#ifdef DIAGNOSE_SDR
	if (write_monitor(hfire->sd_id_fire)) return error(ERR_HDF_QUIET);
#endif

	counts->NumPix = hinfo->nscans * LINES_PER_SCAN_750M * SAMPLES_750M;

	st = write_fire_attributes(hfire->sd_id_fire, counts);
	if (st) return st;

#ifdef NOT_IMP
	if (write_coarse)
		if ( (st = write_fire_attributes(hfire->sd_id_coarse, counts)) )
			return st;
#endif

	return 0;
}




/* fill all fields of specified scan with missing data values */
static void clear_scan(Scan *scan)
/*
!C*****************************************************************************
!Description:
	Fill specified scan buffer with missing data values.

!Input Parameters:
	None.

!Output Parameters:
	scan	pointer to scan

!Input/Output Parameters:
	None.

!Returns:
	Nothing.

!Revision History
	Original version January 2001, L. Giglio.

!Team-unique Header:

!References and Credits

!Design Notes

!END***************************************************************************
*/
{
	int i, j;

	/* missing data value used for radiance, reflectance, and brightness
	temperature fields */
	const float32 missval = -1.0F;

	const float missgeo = -999.0F;

	PixInfo NullPixInfo;


	(void) memset(&NullPixInfo, 0, sizeof(PixInfo));

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++) {
			scan->T13[i][j] = scan->T15[i][j] =
				scan->T16[i][j] = missval;
			scan->R5[i][j] = scan->R7[i][j] =
				scan->R7[i][j] = missval;
			scan->L13[i][j] = missval;

			scan->QF5[i][j] = scan->QF7[i][j] =
				scan->QF11[i][j] = 0;
			scan->QF13[i][j] = scan->QF15[i][j] =
				scan->QF16[i][j] = 0;

			scan->lat[i][j] = scan->lon[i][j] = missgeo;

			scan->SolZenAng[i][j] = scan->ViewZenAng[i][j] =
				scan->SolAzAng[i][j] = scan->ViewAzAng[i][j] =
				missgeo;

			scan->landsea[i][j] = 0;

			scan->pixflags[i][j] = NullPixInfo;
			}
		}
}



static void update_cmg(OutScan *outscan, Scan *scan)
{
	int i, j;

	for (i = 0; i < LINES_PER_SCAN_750M; i++) {
		for (j = 0; j < SAMPLES_750M; j++)
			cmg_append(scan->lat[i][j], scan->lon[i][j],
				outscan->ExtFireMask[i][j],
				scan->pixflags[i][j]);
		}
}



