/*
Functions for checking QF1 state of SDR input bands used by
the detection algorithm.
*/

#include "mfhdf.h"
#include "qf.h"


/*
Order of explanatory text below:
quality / pre-agg. saturation / missing data / data range

Bit-field order:
data range | missing data | pre-agg. saturation | quality

Usable data:

     5 0|0|1|1	poor / some / ok / ok
     9 0|0|2|1	poor / all / ok / ok

    33 0|2|0|1	poor / none / missing cal. / ok
    65 1|0|0|1	poor / none / ok / radiance out of range

   129 2|0|0|1	poor / none / ok / R or EBBT out of range
   161 2|2|0|1	poor / none / missing cal. / R or EBBT out of range
   165 2|2|1|1	poor / some / ok / R or EBBT out of range
   133 2|0|1|1	poor / some / ok / R or EBBT out of range
   193 3|0|0|1	poor / none / ok / radiance and EBBT out of range

Treat everything else as unusable.  Here's a sample:

     2 0|0|0|2	(trimmed pixel)
    34 0|2|0|2	no. cal / none / missing cal. / ok
    50 0|3|0|2	no. cal / none / thermistor data missing / ok
    18 0|1|0|2	no. cal / none / EV RDR data missing / ok

*/



/*
Return Value
	0	valid data
	1	invalid and/or poor quality data
*/
int test_therm_QF(int bandnum, uint8 QF1, float Tb)
{
	int quality;		/* bits 0-1 */
#ifdef UNUSED
	int saturation;		/* bits 2-3 */
	int b3, b4;
#endif

	/* missing thermal values are always (day + night) bad */
	if (Tb <= 0.0F) return 1;


	/*
	As of Mx8.4, it appears that the SDR QF1 bits are being set
	correctly, so we can eliminate the special tests we originally used to
	let specific (incorrect) QF1 values pass through.
	This simplifies the subsequent code considerably.
	*/

	quality = QF1 & 3;
#ifdef UNUSED
	saturation = (QF1 >> 2) & 3;
	b3 = (QF1 >> 4) & 3;
	b4 = (QF1 >> 6) & 3;
#endif

	/*
	Permit good quality (0) and poor quality (1) pixels.
	Disallow pixels having no calibration (2) or an invalid quality
	value (3).
	*/
	return quality > 1;
}


