/*
!Description:

	Implementation of alternate VIIRS active fire detection
	algorithm described in forthcoming paper by Giglio, et al.

!References and Credits:

	L. Giglio, University of Maryland
	W. Schroeder, University of Maryland
	I. Csiszar, NOAA/NESDIS
	C. O. Justice, University of Maryland

	Baseline version, May 2012, L. Giglio.

!Design Notes:

	This code is a port of the MODIS Collection 6 swath-level fire product
	production code with VIIRS-specific extensions.  It will continue to
	be modified over time to better exploit the capabilities and
	idiosyncrasies of the VIIRS sensor.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#include "mfhdf.h"


#include "const.h"
#include "types.h"
#include "fire.h"
#include "output.h"
#include "input.h"
#include "process.h"
#include "firetab.h"
#include "cmg.h"
#include "error.h"
#include "version.h"







int verbose;		/* flag: verbose mode */
int write_cmg;		/* flag: append CMG input-data layer to product */
int write_tb;		/* flag: append brightness temp. SDSs to product */
int write_ref;		/* flag: append reflectance SDSs to product */
int compress_sds;	/* flag: use internal HDF compression to compress 2D SDSs */
int write_geo;		/* flag: append geolocation SDSs to product */





int main(int argc, char **argv)
{
	int32 sd_vmae;		/* NPP_VMAE_L1 */
	int32 sd_lc;		/* NPP_QSLWMIP_L2 */
	int32 sd_fire;

	char *infile_vmae;	/* M-band SDR product */
	char *infile_lc;	/* granulated quarterly surface type */

	HDF_InputInfo hin;
	HDF_OutputInfo hout;

	char *outfile;
	int st;

	char msg[MAX_MSG_LEN];

	FireCounts counts;

	int skip_water;		/* flag: do not process water pixels */

	int c;
	extern char *optarg;
	extern int optind, opterr;

#ifdef DEBUG
printf("sizeof(PixInfo) = %d\n", sizeof(PixInfo));
#endif


	verbose = write_cmg = compress_sds = 0;
	write_tb = write_ref = write_geo = 0;
	skip_water = 0;

	while ((c = getopt(argc, argv, "gLrtvw")) != -1) {
		switch (c) {
			case 'v':
				verbose = -1;
				break;

			case 'L':
				write_geo = -1;
				break;

			case 'r':
				write_ref = -1;
				break;

			case 't':
				write_tb = -1;
				break;

			case 'g':
				write_cmg = -1;
				break;

			case 'w':
				skip_water = -1;
				break;

			default:
				USAGE:
				fputs("Usage: vfire [-gLrtvw] VMAE_L1_file "
					"QSLWMIP_L2_file output_file\n",
					stderr);
				fputs("    g    Append CMG input data SDS to output product.\n", stderr);
				fputs("    L    Add geolocation SDSs to product.\n", stderr);
				fputs("    r    Append M7 and M11 reflectance SDSs to output product.\n", stderr);
				fputs("    t    Append M13 and M15 brightness temperature SDSs to product.\n", stderr);
				fputs("    v    Verbose mode.\n", stderr);
				fputs("    w    Skip water pixels during processing (a la MODIS C5 and earlier).\n", stderr);
				fprintf(stderr, "\nVersion %s, compiled %s %s.\n",
					PROCESS_VERSION_NUMBER, __DATE__, __TIME__);
#ifdef DIAGNOSE_SDR
	fputs("Compiled with DIAGNOSE_SDR.\n", stderr);
#endif
				exit(1);
				break;
			}
		}


	/* make sure three file names follow any switches */
	if (optind != argc - 3) goto USAGE;

	infile_vmae = argv[optind];
	infile_lc = argv[optind+1];
	outfile = argv[optind+2];


	if (verbose) printf("%s\n%s\n%s\n", infile_vmae, infile_lc, outfile);


	if ((sd_vmae = SDstart(infile_vmae, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".",
			infile_vmae);
		errmsg(msg);
		exit(1);
		}

	if ((sd_lc = SDstart(infile_lc, DFACC_READ)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF input file \"%s\".",
			infile_lc);
		errmsg(msg);
		exit(1);
		}

	if ( (st = InitInputHDF(sd_vmae, sd_lc, &hin)) ) exit(st);

	if (verbose) printf("nscans = %d\n", hin.nscans);


	if ((sd_fire = SDstart(outfile, DFACC_CREATE)) == FAIL) {
		(void) error(ERR_HDF_QUIET);
		(void) sprintf(msg, "Unable to open HDF output file \"%s\".", outfile);
		errmsg(msg);
		exit(3);
		}


	st = InitOutputHDF(sd_fire, hin.nscans, write_geo, &hout);
	if (st) exit(4);


	if (cmg_init(write_cmg)) {
		exit(4);
		}

	st = ProcGranule(&hin, &hout, &counts, skip_water);
	if (st) {
		fputs("Error processing granule.\n", stderr);
		exit(3);
		}

	if (verbose) {
		printf("Fire:       %7d\n", counts.NumFirePix);
		printf("Unknown:    %7d\n", counts.NumUnknownLandPix);
		}

	if (write_cmg) {
if (verbose) puts("Writing CMG layers...");
		if (cmg_write(hout.sd_id_fire, CMG_DAY)) exit(5);
		if (cmg_write(hout.sd_id_fire, CMG_NIGHT)) exit(5);
		}


	if (write_generic_attributes(hout.sd_id_fire, infile_vmae,
		infile_lc, hin.sat, hin.gringLatLonStr)) exit(5);

	if (SDend(sd_fire) == FAIL) exit(error(ERR_HDF));

	if (SDend(sd_lc) == FAIL) exit(error(ERR_HDF));
	if (SDend(sd_vmae) == FAIL) exit(error(ERR_HDF));

	free(hin.gringLatLonStr[0]);
	free(hin.gringLatLonStr[1]);

	free_fire_pixel_table();

	cmg_free(write_cmg);


	exit(0);
}


